#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from functools import lru_cache
from typing import Optional

KHMER_CALENDAR = "KHMER_CALENDAR"
THAI_CALENDAR = "THAI_CALENDAR"


class _ThaiLunisolar:
    """
    ** Thai Lunar Calendar Holidays only work from 1941 (B.E. 2484) onwards
       until 2057 (B.E. 2600) as we only have Thai year-type data for
       cross-checking until then.

    So here are the basics of the Thai Lunar Calendar
        3-year types for calendar intercalation:
            - Pakatimat (Normal Year):
                    consist of 12 months, has 354 days.
            - Athikawan (Extra-Day Year):
                    add a day to the 7th month of the year, has 355 days
                    for the synodic month correction.
            - Athikamat (Extra-Month Year):
                    we have the 8th month twice, has 384 days for the
                    sidereal year correction.

        Each month either has 30 (Even months) or 29 (Odd months)
            - The waxing phase has 15 days until Full Moon and waning
                phase 14 (Odd Months)/15 (Even Months/
                Month 7 of Athikawan years) days for the New Moon.
            - The second "Month 8" for Athikamat years is called
                "Month 8.8", with all observed holy days delayed from
                the usual calendar by 1 month.

    List of public holidays dependent on the Thai Lunar Calendar:
        - Magha Puja/Makha Bucha/Meak Bochea:
                15th Waxing Day (Full Moon) of Month 3
                (On Month 4 for Athikamat Years).
                KHMER_CALENDAR always fall on Month 3.
        - Vesak/Visakha Bucha/Visaka Bochea:
                15th Waxing Day (Full Moon) of Month 6
                (On Month 7 for Athikamat Years).
                KHMER_CALENDAR always fall on Month 6.
        - Thai Royal Ploughing Ceremony/Raeknakhwan:
                Based on this, though Court Astrologer picks the
                auspicious dates, which sadly don't fall into a
                predictable pattern; see its specific section below.
        - Cambodian Royal Ploughing Ceremony/Preah Neangkol:
                4th Waning Day of Month 6
                (On Month 7 for Athikamat Years).
                This defaults to KHMER_CALENDAR (its sole user).
        - Asalha Puja/Asarnha Bucha:
                15th Waxing Day (Full Moon) of Month 8
                (On Month 8/8 for Athikamat Years).
                KHMER_CALENDAR always fall on Month 8.
        - Buddhist Lent Day/Wan Khao Phansa:
                1st Waning Day of Month 8
                (On Month 8/8 for Athikamat Years).
                KHMER_CALENDAR always fall on Month 8.
        - Pchum Ben/Prachum Bandar:
                15th Waning Day (New Moon) of Month 10.
        - Loy Krathong/Boun That Louang/Bon Om Touk:
                15th Waxing Day (Full Moon) of Month 12.

    Other Buddhist date on Thai Lunar Calendar:
        - Buddha's Cremation Day/Atthami Bucha
                8th Waning Day of  Month 6
                (On Month 7 for Athikamat Years).
                KHMER_CALENDAR always fall on Month 6
        - End of Buddhist Lent Day/Ok Phansa:
                15th Waxing Day (Full Moon) of Month 11

    The following code is based on Ninenik Narkdee's PHP implementation,
    and we're thankful for his work.

    Please avoid touching the Athikawan and Athikamat declaration array
    at all costs unless you can find sources for them somewhere for 2057++

    Sources: (Ninenik.com 's wbm) http://tiny.cc/wa_ninenik_thluncal_php
             https://www.myhora.com/ปฏิทิน/ปฏิทิน-พ.ศ.2560.aspx

    Usage example:

    >>> from holidays.utils import _ThaiLunisolar
    >>> thls = _ThaiLunisolar()
    >>> print(thls.visakha_bucha_date(2010))
    2010-05-28
    """

    # Athikawan (Extra-Day Year) list goes from 1941-2057 C.E.
    # Copied off from 1757-2057 (B.E. 2300-2600) Thai Lunar Calendar
    ATHIKAWAN_YEARS_GREGORIAN = {
        1945,
        1949,
        1952,
        1957,
        1963,
        1970,
        1973,
        1979,
        1987,
        1990,
        1997,
        2000,
        2006,
        2009,
        2016,
        2020,
        2025,
        2032,
        2035,
        2043,
        2046,
        2052,
    }

    # Athikamat (Extra-Month Year) list goes from 1941-2057 C.E.:
    # Copied off from 1757-2057 (B.E. 2300-2600) Thai Lunar Calendar
    # Approx formula as follows: (common_era-78)-0.45222)%2.7118886 < 1
    ATHIKAMAT_YEARS_GREGORIAN = {
        1942,
        1944,
        1947,
        1950,
        1953,
        1956,
        1958,
        1961,
        1964,
        1966,
        1969,
        1972,
        1975,
        1977,
        1980,
        1983,
        1985,
        1988,
        1991,
        1994,
        1996,
        1999,
        2002,
        2004,
        2007,
        2010,
        2012,
        2015,
        2018,
        2021,
        2023,
        2026,
        2029,
        2031,
        2034,
        2037,
        2040,
        2042,
        2045,
        2048,
        2050,
        2053,
        2056,
    }

    # While Buddhist Holy Days have been observed since the 1900s
    #   Due to the calendar changes in 1941 (B.E. 2484) and that
    #   our array only goes up to B.E. 2600; We'll thus only populate
    #   the data for 1941-2057 (B.E. 2484-2600).
    # Sources: หนังสือเวียนกรมการปกครอง กระทรวงมหาดไทย มท 0310.1/ว4 5 ก.พ. 2539
    START_DATE = date(1940, 11, 30)
    START_YEAR = 1941
    END_YEAR = 2057

    def __init__(self, calendar=THAI_CALENDAR) -> None:
        self.__verify_calendar(calendar)
        self.__calendar = calendar

    @staticmethod
    def __is_khmer_calendar(calendar):
        """
        Return True if `calendar` is Khmer calendar.
        Return False otherwise.
        """
        return calendar == KHMER_CALENDAR

    @staticmethod
    def __verify_calendar(calendar):
        """
        Verify calendar type.
        """
        if calendar not in {KHMER_CALENDAR, THAI_CALENDAR}:
            raise ValueError(
                f"Unknown calendar name: {calendar}. Use `KHMER_CALENDAR` or `THAI_CALENDAR`."
            )

    @lru_cache()
    def _get_start_date(self, year: int) -> Optional[date]:
        """
        Calculate the start date of that particular Thai Lunar Calendar Year.
        This usually falls in November or December of the previous Gregorian
        year in question. Should the year be outside of working scope
        (1941-2057: B.E 2484-2600), this will returns None instead.

        :param year:
            The Gregorian year.

        :return:
             The start date of Thai Lunar Calendar for a Gregorian year.
        """
        if year < _ThaiLunisolar.START_YEAR or year > _ThaiLunisolar.END_YEAR:
            return None

        iter_start_date = _ThaiLunisolar.START_DATE
        iter_start_year = _ThaiLunisolar.START_YEAR

        while iter_start_year < year:
            if iter_start_year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN:
                delta_days = 384
            elif iter_start_year in _ThaiLunisolar.ATHIKAWAN_YEARS_GREGORIAN:
                delta_days = 355
            else:
                delta_days = 354
            iter_start_date += td(days=delta_days)
            iter_start_year += 1
        return iter_start_date

    def makha_bucha_date(self, year: int, calendar=None) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Makha Bucha.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "Magha Puja", "Makha Buxha" and "Meak Bochea".
        This concides with the 15th Waxing Day of Month 3
        in Thai Lunar Calendar, or Month 4 in Athikamat years.

        KHMER_CALENDAR will always use Month 3 regardless of year type.

        To calculate, we use use the following time delta:
        - Athikamat: 15th Waxing Day of Month 4
                     or 29[1] + 30[2] + 29[3] + 15[4] -1 = 102
        - Athikawan: 15th Waxing Day of Month 3
                     or 29[1] + 30[2] + 15[3] -1 = 73
        - Pakatimat: 15th Waxing Day of Month 3
                     or 29[1] + 30[2] + 15[3] -1 = 73

        :param year:
            The Gregorian year.

        :param calendar:
            Calendar type, this defaults to THAI_CALENDAR.

        :return:
            Estimated Gregorian date of Makha Bucha.
        """
        calendar = calendar or self.__calendar
        self.__verify_calendar(calendar)

        start_date = self._get_start_date(year)
        if not start_date:
            return None

        return start_date + td(
            days=+102
            if (
                year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN
                and not self.__is_khmer_calendar(calendar)
            )
            else +73
        )

    def visakha_bucha_date(self, year: int, calendar=None) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Visakha Bucha.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "Vesak" and "Buddha Day". This concides with
        the 15th Waxing Day of Month 6 in Thai Lunar Calendar,
        or Month 7 in Athikamat years.

        KHMER_CALENDAR will always use Month 6 regardless of year type.

        To calculate, we use use the following time delta:
        - Athikamat: 15th Waxing Day of Month 7
                     or 177[1-6] + 15[7] -1 = 191
        - Athikawan: 15th Waxing Day of Month 6
                     or 147[1-5] + 15[6] -1 = 161
        - Pakatimat: 15th Waxing Day of Month 6
                     or 147[1-5] + 15[6] -1 = 161

        :param year:
            The Gregorian year.

        :param calendar:
            Calendar type, this defaults to THAI_CALENDAR.

        :return:
            Estimated Gregorian date of Visakha Bucha.
        """
        calendar = calendar or self.__calendar
        self.__verify_calendar(calendar)

        start_date = self._get_start_date(year)
        if not start_date:
            return None

        return start_date + td(
            days=+191
            if (
                year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN
                and not self.__is_khmer_calendar(calendar)
            )
            else +161
        )

    def preah_neangkoal_date(self, year: int) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Preah Neangkoal.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "Cambodian Royal Ploughing Ceremony". This always
        concides with the 4th Waning Day of Month 6 in Khmer Lunar Calendar.

        To calculate, we use use the following time delta:
        - Athikamat: 15th Waxing Day of Month 6
                     or 177[1-6] + 19[7] -1 = 165
        - Athikawan: 15th Waxing Day of Month 6
                     or 147[1-5] + 19[6] -1 = 165
        - Pakatimat: 15th Waxing Day of Month 6
                     or 147[1-5] + 19[6] -1 = 165
        Or as in simpler terms: "Visakha Bucha" +4

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Preah Neangkoal.
        """
        start_date = self._get_start_date(year)
        if not start_date:
            return None

        return start_date + td(days=+165)

    def atthami_bucha_date(self, year: int, calendar=None) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Atthami Bucha.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "Buddha's Cremation Day". This concides with
        the 8th Waning Day of Month 6 in Thai Lunar Calendar,
        or Month 7 in Athikamat years.

        KHMER_CALENDAR will always use Month 6 regardless of year type.

        To calculate, we use use the following time delta:
        - Athikamat: 8th Waning Day of  Month 7
                    or 177[1-6] + 23[7] -1 = 199
        - Athikawan: 8th Waning Day of  Month 6
                     or 147[1-5] + 23[6] -1 = 169
        - Pakatimat: 8th Waning Day of  Month 6
                    or 147[1-5] + 23[6] -1 = 169
        - Or as in simpler terms: "Visakha Bucha" +8

        :param year:
            The Gregorian year.

        :param calendar:
            Calendar type, this defaults to THAI_CALENDAR.

        :return:
            Estimated Gregorian date of Atthami Bucha.
        """
        calendar = calendar or self.__calendar
        self.__verify_calendar(calendar)

        start_date = self._get_start_date(year)
        if not start_date:
            return None

        return start_date + td(
            days=+199
            if (
                year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN
                and not self.__is_khmer_calendar(calendar)
            )
            else +169
        )

    def asarnha_bucha_date(self, year: int, calendar=None) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Asarnha Bucha.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "Asalha Puja". This concides with
        the 15th Waxing Day of Month 8 in Thai Lunar Calendar,
        or Month 8.8 in Athikamat years.

        KHMER_CALENDAR will always use Month 8 regardless of year type.

        To calculate, we use use the following time delta:
        - Athikamat: 15th Waxing Day of Month 8/8
                     or 177[1-6] + 29[7] + 30[8] + 15[8.8] -1 = 250
        - Athikawan: 15th Waxing Day of Month 8
                     or 177[1-6] + 30[7] + 15[8] -1 = 221
        - Pakatimat: 15th Waxing Day of Month 8
                     or 177[1-6] + 29[7] + 15[8] -1 = 220

        :param year:
            The Gregorian year.

        :param calendar:
            Calendar type, this defaults to THAI_CALENDAR.

        :return:
            Estimated Gregorian date of Asarnha Bucha.
        """
        calendar = calendar or self.__calendar
        self.__verify_calendar(calendar)

        start_date = self._get_start_date(year)
        if not start_date:
            return None

        if year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN and not self.__is_khmer_calendar(
            calendar
        ):
            delta_days = 250
        elif year in _ThaiLunisolar.ATHIKAWAN_YEARS_GREGORIAN:
            delta_days = 221
        else:
            delta_days = 220
        return start_date + td(days=delta_days)

    def khao_phansa_date(self, year: int, calendar=None) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Khao Phansa.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "(Start of) Buddhist Lent" and "Start of Vassa".
        This concides with the 1st Waning Day of Month 8
        in Thai Lunar Calendar, or Month 8.8 in Athikamat years.

        KHMER_CALENDAR will always use Month 8 regardless of year type.

        To calculate, we use use the following time delta:
        - Athikamat: 1st Waning Day of Month 8.8
                     or 177[1-6] + 29[7] + 30[8] + 16[8.8] -1 = 251
        - Athikawan: 1st Waning Day of Month 8 ]
                     or 177[1-6] + 30[7] + 16[8] -1 = 222
        - Pakatimat: 1st Waning Day of Month 8
                     or 177[1-6] + 29[7] + 16[8] -1 = 221
        - Or as in simpler terms: "Asarnha Bucha" +1

        :param year:
            The Gregorian year.

        :param calendar:
            Calendar type, this defaults to THAI_CALENDAR.

        :return:
            Estimated Gregorian date of Khao Phansa.
        """
        calendar = calendar or self.__calendar
        self.__verify_calendar(calendar)

        start_date = self._get_start_date(year)
        if not start_date:
            return None

        if year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN and not self.__is_khmer_calendar(
            calendar
        ):
            delta_days = 251
        elif year in _ThaiLunisolar.ATHIKAWAN_YEARS_GREGORIAN:
            delta_days = 222
        else:
            delta_days = 221
        return start_date + td(days=delta_days)

    def pchum_ben_date(self, year: int) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Pchum Ben.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "Prachum Bandar".
        This concides with the 15th Waning Day of Month 10 in
        Thai Lunar Calendar.

         To calculate, we use use the following time delta:
        - Athikamat: 15th Waning Day of Month 10
                     or 265[1-9] + 30[8.8] + 30[10] -1 = 324
        - Athikawan: 15th Waning Day of Month 10
                     or 265[1-9] + 1[7] + 30[10] -1 = 295
        - Pakatimat: 15th Waning Day of Month 10
                     or 265[1-9] + 30[10] -1 = 294

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Pchum Ben.
        """
        start_date = self._get_start_date(year)
        if not start_date:
            return None

        if year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN:
            delta_days = 324
        elif year in _ThaiLunisolar.ATHIKAWAN_YEARS_GREGORIAN:
            delta_days = 295
        else:
            delta_days = 294
        return start_date + td(days=delta_days)

    def ok_phansa_date(self, year: int) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Ok Phansa.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "End of Buddhist Lent" and "End of Vassa".
        This concides with the 15th Waxing Day of Month 11
        in Thai Lunar Calendar.

        To calculate, we use use the following time delta:
        - Athikamat: 15th Waxing Day of Month 11
                     or 295[1-10] + 30[8.8] + 15[11] -1 = 339
        - Athikawan: 15th Waxing Day of Month 11
                     or 295[1-10] + 1[7] + 15[11] -1 = 310
        - Pakatimat: 15th Waxing Day of Month 11
                     or 295[1-10] + 15[11] -1 = 309

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Ok Phansa.
        """
        start_date = self._get_start_date(year)
        if not start_date:
            return None

        if year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN:
            delta_days = 339
        elif year in _ThaiLunisolar.ATHIKAWAN_YEARS_GREGORIAN:
            delta_days = 310
        else:
            delta_days = 309
        return start_date + td(days=delta_days)

    def loy_krathong_date(self, year: int) -> Optional[date]:
        """
        Calculate the estimated Gregorian date of Loy Krathong.
        If the Gregorian year input is invalid, this will outputs None instead.

        Also known as "Boun That Louang" and "Bon Om Touk".
        This concides with the 15th Waxing Day of Month 12
        in Thai Lunar Calendar.

        To calculate, we use use the following time delta:
        - Athikamat: 15th Waxing Day of Month 12
                     or 324[1-11] + 30[8.8] + 15[11] -1 = 368
        - Athikawan: 15th Waxing Day of Month 12
                     or 324[1-11] + 1[7] + 15[11] -1 = 339
        - Pakatimat: 15th Waxing Day of Month 12
                     or 324[1-11] + 15[11] -1 = 338

        :param year:
            The Gregorian year.

        :return:
            Estimated Gregorian date of Loy Krathong.
        """
        start_date = self._get_start_date(year)
        if not start_date:
            return None

        if year in _ThaiLunisolar.ATHIKAMAT_YEARS_GREGORIAN:
            delta_days = 368
        elif year in _ThaiLunisolar.ATHIKAWAN_YEARS_GREGORIAN:
            delta_days = 339
        else:
            delta_days = 338
        return start_date + td(days=delta_days)



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars.gregorian import MAR, JUL, AUG, SEP, FRI, SAT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Andorra(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Andorra
      - https://www.holsdb.com/public-holidays/ad
    """

    country = "AD"
    subdivisions = (
        "02",  # Canillo.
        "03",  # Encamp.
        "04",  # La Massana.
        "05",  # Ordino.
        "06",  # Sant Julià de Lòria.
        "07",  # Andorra la Vella.
        "08",  # Escaldes-Engordany.
    )

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year: int) -> None:
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Epiphany.
        self._add_epiphany_day("Epiphany")

        # Carnival.
        self._add_carnival_tuesday("Carnival")

        # Constitution Day.
        self._add_holiday("Constitution Day", MAR, 14)

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Easter Sunday.
        self._add_easter_monday("Easter Monday")

        # Labor Day.
        self._add_labor_day("Labor Day")

        # Whit Monday.
        self._add_whit_monday("Whit Monday")

        # Assumption Day.
        self._add_assumption_of_mary_day("Assumption Day")

        # National Day.
        self._add_holiday("National Day", SEP, 8)

        # All Saints' Day.
        self._add_all_saints_day("All Saints' Day")

        # Immaculate Conception Day.
        self._add_immaculate_conception_day("Immaculate Conception Day")

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Saint Stephen's Day.
        self._add_christmas_day_two("Saint Stephen's Day")

    # Canillo.
    def _add_subdiv_02_holidays(self):
        name = "Canillo Annual Festival"
        third_sat_of_july = self._get_nth_weekday_of_month(3, SAT, JUL)
        self._add_holiday(name, third_sat_of_july)
        self._add_holiday(name, third_sat_of_july + td(days=+1))
        self._add_holiday(name, third_sat_of_july + td(days=+2))

    # Encamp.
    def _add_subdiv_03_holidays(self):
        name = "Encamp Annual Festival"
        aug_15 = self._add_holiday(name, AUG, 15)
        self._add_holiday(name, aug_15 + td(days=+1))

    # La Massana.
    def _add_subdiv_04_holidays(self):
        name = "La Massana Annual Festival"
        aug_15 = self._add_holiday(name, AUG, 15)
        self._add_holiday(name, aug_15 + td(days=+1))

    # Ordino.
    def _add_subdiv_05_holidays(self):
        name = "Ordino Annual Festival"
        aug_15 = self._add_holiday(name, AUG, 15)
        self._add_holiday(name, aug_15 + td(days=+1))

    # Sant Julià de Lòria.
    def _add_subdiv_06_holidays(self):
        name = "Sant Julià de Lòria Annual Festival"
        last_fri_of_july = self._get_nth_weekday_from(-1, FRI, JUL, 29)
        self._add_holiday(name, last_fri_of_july)
        self._add_holiday(name, last_fri_of_july + td(days=+1))
        self._add_holiday(name, last_fri_of_july + td(days=+2))
        self._add_holiday(name, last_fri_of_july + td(days=+3))

    # Andorra la Vella.
    def _add_subdiv_07_holidays(self):
        name = "Andorra la Vella Annual Festival"
        first_sat_of_august = self._get_nth_weekday_of_month(1, SAT, AUG)
        self._add_holiday(name, first_sat_of_august)
        self._add_holiday(name, first_sat_of_august + td(days=+1))
        self._add_holiday(name, first_sat_of_august + td(days=+2))

    # Escaldes-Engordany.
    def _add_subdiv_08_holidays(self):
        name = "Escaldes-Engordany Annual Festival"
        jul_25 = self._add_holiday(name, JUL, 25)
        self._add_holiday(name, jul_25 + td(days=+1))


class AD(Andorra):
    pass


class AND(Andorra):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import MAY, AUG, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Ecuador(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Ecuador
      - http://tiny.cc/ec_co_tr
    """

    country = "EC"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(
        self, dt: date, weekend_only: bool = False, before: bool = True, after: bool = True
    ) -> None:
        if self.observed and self._year >= 2017:
            obs_date = None
            # Art. 1 of Law #0 from 20.12.2016
            # When holidays falls on Tuesday, the rest shall be transferred to
            # preceding Monday, and if they falls on Wednesday or Thursday,
            # the rest shall be transferred to Friday of the same week.
            # Exceptions to this provision are January 1, December 25 and
            # Shrove Tuesday.
            if not weekend_only:
                if self._is_tuesday(dt) and before:
                    obs_date = dt + td(days=-1)
                elif self._is_wednesday(dt):
                    obs_date = dt + td(days=+2)
                elif self._is_thursday(dt) and after:
                    obs_date = dt + td(days=+1)
            # When holidays falls on Saturday or Sunday, the rest shall be
            # transferred, respectively, to the preceding Friday or the
            # following Monday.
            if self._is_saturday(dt) and before:
                obs_date = dt + td(days=-1)
            elif self._is_sunday(dt) and after:
                obs_date = dt + td(days=+1)
            if obs_date:
                self._add_holiday(self.tr("%s (Observado)") % self[dt], obs_date)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        name = self.tr("Año Nuevo")
        self._add_observed(self._add_new_years_day(name), weekend_only=True)

        if self.observed and year >= 2017:
            if self._is_friday(DEC, 31):
                self._add_holiday(self.tr("%s (Observado)") % name, DEC, 31)

        # Carnival.
        name = tr("Carnaval")
        self._add_carnival_monday(name)
        self._add_carnival_tuesday(name)

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Labour Day.
        self._add_observed(self._add_labor_day(tr("Día del Trabajo")))

        # The Battle of Pichincha.
        self._add_observed(self._add_holiday(tr("Batalla de Pichincha"), MAY, 24))

        # Declaration of Independence of Quito.
        self._add_observed(self._add_holiday(tr("Primer Grito de Independencia"), AUG, 10))

        # Independence of Guayaquil.
        self._add_observed(self._add_holiday(tr("Independencia de Guayaquil"), OCT, 9))

        # All Souls' Day.
        self._add_observed(self._add_all_souls_day(tr("Día de los Difuntos")), after=False)

        # Independence of Cuenca.
        self._add_observed(self._add_holiday(tr("Independencia de Cuenca"), NOV, 3), before=False)

        # Christmas Day.
        self._add_observed(self._add_christmas_day(tr("Día de Navidad")), weekend_only=True)


class EC(Ecuador):
    pass


class ECU(Ecuador):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from gettext import gettext as tr

from holidays.calendars.gregorian import JUN, SEP, OCT, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Guatemala(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - http://www.bvnsa.com.gt/bvnsa/calendario_dias_festivos.php
    - https://www.minfin.gob.gt/images/downloads/leyes_acuerdos/decretocong19_101018.pdf
    """

    country = "GT"
    default_language = "es"
    supported_languages = ("en_US", "es")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _move_holiday(self, dt: date) -> None:
        """
        law 19-2018 start 18 oct 2018
        https://www.minfin.gob.gt/images/downloads/leyes_acuerdos/decretocong19_101018.pdf


        EXPEDIENTE 5536-2018 (CC) start 17 abr 2020
        https://leyes.infile.com/index.php?id=181&id_publicacion=81051
        """
        if dt <= date(2018, OCT, 17) or self._is_monday(dt):
            return None
        self._add_holiday(
            self[dt],
            self._get_nth_weekday_from(-1, MON, dt)
            if self._is_tuesday(dt) or self._is_wednesday(dt)
            else self._get_nth_weekday_from(1, MON, dt),
        )
        self.pop(dt)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Good Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Good Saturday.
        self._add_holy_saturday(tr("Sabado Santo"))

        # Labor Day.
        dt = self._add_labor_day(tr("Dia del Trabajo"))
        if year == 2019:
            self._move_holiday(dt)

        # Army Day.
        self._move_holiday(self._add_holiday(tr("Dia del Ejército"), JUN, 30))

        # Assumption Day.
        self._add_assumption_of_mary_day(tr("Dia de la Asunción"))

        # Independence Day
        self._add_holiday(tr("Día de la Independencia"), SEP, 15)

        # Revolution Day
        dt = self._add_holiday(tr("Dia de la Revolución"), OCT, 20)
        if year in {2018, 2019}:
            self._move_holiday(dt)

        # All Saints' Day.
        self._add_all_saints_day(tr("Dia de Todos los Santos"))

        # Christmas Day.
        self._add_christmas_day(tr("Dia de Navidad"))


class GT(Guatemala):
    pass


class GUA(Guatemala):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from holidays.calendars.gregorian import JAN, MAR, MAY, AUG, SEP, NOV, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Belize(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Belize
      - http://www.belizelaw.org/web/lawadmin/PDF%20files/cap289.pdf
      - https://www.pressoffice.gov.bz/public-and-bank-holidays-2022-updated/
      - https://www.pressoffice.gov.bz/government-of-belize-establishes-new-public-and-bank-holidays/  # noqa: E501
    """

    country = "BZ"

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _move_holiday(self, dt: date, sunday_only: bool = True) -> None:
        # Chapter 289 of the laws of Belize states that if the holiday falls
        # on a Sunday or a Friday, the following Monday is observed as public
        # holiday; further, if the holiday falls on a Tuesday, Wednesday or
        # Thursday, the preceding Monday is observed as public holiday
        if not self.observed:
            return None
        obs_dt = None
        if sunday_only:
            if self._is_sunday(dt):
                obs_dt = self._get_nth_weekday_from(1, MON, dt)
        else:
            if self._is_friday(dt) or self._is_sunday(dt):
                obs_dt = self._get_nth_weekday_from(1, MON, dt)
            elif self._is_tuesday(dt) or self._is_wednesday(dt) or self._is_thursday(dt):
                obs_dt = self._get_nth_weekday_from(-1, MON, dt)
        if obs_dt:
            self._add_holiday("%s (Observed)" % self[dt], obs_dt)
            self.pop(dt)

    def _populate(self, year):
        # Belize was granted independence on 21.09.1981.
        if year <= 1981:
            return None
        super()._populate(year)

        # New Year's Day.
        self._move_holiday(self._add_new_years_day("New Year's Day"))

        if year >= 2021:
            # George Price Day.
            self._move_holiday(self._add_holiday("George Price Day", JAN, 15))

        # National Heroes and Benefactors Day.
        self._move_holiday(
            self._add_holiday("National Heroes and Benefactors Day", MAR, 9), sunday_only=False
        )

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Holy Saturday.
        self._add_holy_saturday("Holy Saturday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Labour Day.
        self._move_holiday(self._add_labor_day("Labour Day"))

        if year <= 2021:
            # Commonwealth Day.
            self._move_holiday(self._add_holiday("Commonwealth Day", MAY, 24), sunday_only=False)

        if year >= 2021:
            # Emancipation Day.
            self._move_holiday(self._add_holiday("Emancipation Day", AUG, 1), sunday_only=False)

        # Saint George's Caye Day.
        self._move_holiday(self._add_holiday("Saint George's Caye Day", SEP, 10))

        # Independence Day.
        self._move_holiday(self._add_holiday("Independence Day", SEP, 21))

        # Indigenous Peoples' Resistance Day / Pan American Day.
        name = "Indigenous Peoples' Resistance Day" if year >= 2021 else "Pan American Day"
        self._move_holiday(self._add_columbus_day(name), sunday_only=False)

        # Garifuna Settlement Day.
        self._move_holiday(self._add_holiday("Garifuna Settlement Day", NOV, 19))

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Boxing Day.
        self._move_holiday(self._add_christmas_day_two("Boxing Day"))


class BZ(Belize):
    pass


class BLZ(Belize):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from typing import Tuple, Optional

from holidays.calendars import _ChineseLunisolar
from holidays.calendars.gregorian import APR


class ChineseCalendarHolidays:
    """
    Chinese lunisolar calendar holidays.
    """

    def __init__(self, calendar=_ChineseLunisolar(), show_estimated=False) -> None:
        self._chinese_calendar = calendar
        self._show_estimated = show_estimated

    @property
    def _chinese_new_year(self):
        """
        Return Chinese New Year date.
        """
        return self._chinese_calendar.lunar_new_year_date(self._year)[0]

    @property
    def _qingming_date(self):
        day = 5
        if (self._year % 4 < 1) or (self._year % 4 < 2 and self._year >= 2009):
            day = 4
        return date(self._year, APR, day)

    @property
    def _mid_autumn_festival(self):
        """
        Return Mid Autumn Festival (15th day of the 8th lunar month) date.
        """
        return self._chinese_calendar.mid_autumn_date(self._year)[0]

    def _add_chinese_calendar_holiday(
        self, name: str, hol_date: Tuple[date, bool], days_delta: int = 0
    ) -> Optional[date]:
        """
        Add Chinese calendar holiday.

        Adds customizable estimation label to holiday name if holiday date
        is an estimation.
        """
        estimated_label = getattr(self, "estimated_label", "%s* (*estimated)")
        dt, is_estimated = hol_date
        if days_delta != 0:
            dt += td(days=days_delta)

        return self._add_holiday(
            self.tr(estimated_label) % self.tr(name)
            if is_estimated and self._show_estimated
            else name,
            dt,
        )

    def _add_chinese_birthday_of_buddha(self, name) -> Optional[date]:
        """
        Add Birthday of the Buddha by Chinese lunar calendar (8th day of the
        4th lunar month).

        Birthday of the Buddha is a Buddhist festival that is celebrated in
        most of East Asia and South Asia commemorating the birth of Gautama
        Buddha, who was the founder of Buddhism.
        https://en.wikipedia.org/wiki/Buddha%27s_Birthday
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.buddha_birthday_date(self._year)
        )

    def _add_chinese_new_years_eve(self, name) -> Optional[date]:
        """
        Add Chinese New Year's Eve (last day of 12th lunar month).

        Chinese New Year's Eve is the day before the Chinese New Year.
        https://en.wikipedia.org/wiki/Chinese_New_Year%27s_Eve
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.lunar_new_year_date(self._year), days_delta=-1
        )

    def _add_chinese_new_years_day(self, name) -> Optional[date]:
        """
        Add Chinese New Year's Day (first day of the first lunar month).

        Chinese New Year is the festival that celebrates the beginning of
        a new year on the traditional lunisolar and solar Chinese calendar.
        https://en.wikipedia.org/wiki/Chinese_New_Year
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.lunar_new_year_date(self._year)
        )

    def _add_chinese_new_years_day_two(self, name) -> Optional[date]:
        """
        Add Chinese New Year's Day Two.

        https://en.wikipedia.org/wiki/Chinese_New_Year
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.lunar_new_year_date(self._year), days_delta=+1
        )

    def _add_chinese_new_years_day_three(self, name) -> Optional[date]:
        """
        Add Chinese New Year's Day Three.

        https://en.wikipedia.org/wiki/Chinese_New_Year
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.lunar_new_year_date(self._year), days_delta=+2
        )

    def _add_chinese_new_years_day_four(self, name) -> Optional[date]:
        """
        Add Chinese New Year's Day Four.

        https://en.wikipedia.org/wiki/Chinese_New_Year
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.lunar_new_year_date(self._year), days_delta=+3
        )

    def _add_chinese_new_years_day_five(self, name) -> Optional[date]:
        """
        Add Chinese New Year's Day Five.

        https://en.wikipedia.org/wiki/Chinese_New_Year
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.lunar_new_year_date(self._year), days_delta=+4
        )

    def _add_qingming_festival(self, name) -> date:
        """
        Add Qingming Festival (15th day after the Spring Equinox).

        The Qingming festival or Ching Ming Festival, also known as
        Tomb-Sweeping Day in English, is a traditional Chinese festival.
        https://en.wikipedia.org/wiki/Qingming_Festival
        """
        return self._add_holiday(name, self._qingming_date)

    def _add_double_ninth_festival(self, name) -> Optional[date]:
        """
        Add Double Ninth Festival (9th day of 9th lunar month).

        The Double Ninth Festival (Chongyang Festival in Mainland China
        and Taiwan or Chung Yeung Festival in Hong Kong and Macau).
        https://en.wikipedia.org/wiki/Double_Ninth_Festival
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.double_ninth_date(self._year)
        )

    def _add_dragon_boat_festival(self, name) -> Optional[date]:
        """
        Add Dragon Boat Festival (5th day of 5th lunar month).

        The Dragon Boat Festival is a traditional Chinese holiday which occurs
        on the fifth day of the fifth month of the Chinese calendar.
        https://en.wikipedia.org/wiki/Dragon_Boat_Festival
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.dragon_boat_date(self._year)
        )

    def _add_hung_kings_day(self, name) -> Optional[date]:
        """
        Add Hùng Kings' Temple Festival (10th day of the 3rd lunar month).

        Vietnamese festival held annually from the 8th to the 11th day of the
        3rd lunar month in honour of the Hùng Kings.
        https://en.wikipedia.org/wiki/H%C3%B9ng_Kings%27_Festival
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.hung_kings_date(self._year)
        )

    def _add_mid_autumn_festival(self, name) -> Optional[date]:
        """
        Add Mid Autumn Festival (15th day of the 8th lunar month).

        The Mid-Autumn Festival, also known as the Moon Festival or
        Mooncake Festival.
        https://en.wikipedia.org/wiki/Mid-Autumn_Festival
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.mid_autumn_date(self._year)
        )

    def _add_mid_autumn_festival_day_two(self, name) -> Optional[date]:
        """
        Add Mid Autumn Festival Day Two (16th day of the 8th lunar month).

        The Mid-Autumn Festival, also known as the Moon Festival or
        Mooncake Festival.
        https://en.wikipedia.org/wiki/Mid-Autumn_Festival
        """
        return self._add_chinese_calendar_holiday(
            name, self._chinese_calendar.mid_autumn_date(self._year), days_delta=+1
        )



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import APR, JUN, JUL, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Venezuela(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://dias-festivos.eu/dias-festivos/venezuela/#
    """

    country = "VE"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        """
        Overview: https://dias-festivos.eu/dias-festivos/venezuela/#
        Various decrees about holidays:
          1909 (AUG 5): https://bit.ly/3J0mWKQ
          1918 (MAY 19): https://bit.ly/3B8O1Jz
          1921 (JUN 11): https://bit.ly/3aUE2gz
          1971 (JUN 22): https://bit.ly/3yZaUN9
          2002 (OCT 10): https://bit.ly/3B7nRqC
          2012 (MAY 7): https://bit.ly/2MT5x97
        """

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Monday of Carnival.
        self._add_carnival_monday(tr("Lunes de Carnaval"))

        # Tuesday of Carnival.
        self._add_carnival_tuesday(tr("Martes de Carnaval"))

        # Maundy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Note: not sure about the start year, but this event happened in 1811
        if year >= 1811:
            # Declaration of Independence.
            self._add_holiday(tr("Declaración de la Independencia"), APR, 19)

        # https://bit.ly/3B4Xd1L
        if year >= 1946:
            # International Worker's Day.
            self._add_labor_day(tr("Dia Mundial del Trabajador"))

        # Note: not sure about the start year, but this event happened in 1824
        if year >= 1971 or 1824 <= year <= 1917:
            # Battle of Carabobo.
            self._add_holiday(tr("Batalla de Carabobo"), JUN, 24)

        # Note: not sure about the start year, but this event happened in 1811
        if year >= 1811:
            # Independence Day.
            self._add_holiday(tr("Día de la Independencia"), JUL, 5)

        if year >= 1918:
            # Birthday of Simon Bolivar.
            self._add_holiday(tr("Natalicio de Simón Bolívar"), JUL, 24)

        if year >= 1921:
            self._add_columbus_day(
                # Day of Indigenous Resistance.
                tr("Día de la Resistencia Indígena")
                if year >= 2002
                # Columbus Day.
                else tr("Día de la Raza")
            )

        # Note: not sure about the start year nor the reason this was
        # Note: celebrated; the historical records are unclear
        if 1909 <= year <= 1917:
            # Unknown Holiday.
            self._add_holiday(tr("Día Festivo Desconocido"), OCT, 28)

        # Christmas Eve.
        self._add_christmas_eve(tr("Nochebuena"))

        # Christmas Day.
        self._add_christmas_day(tr("Día de Navidad"))

        # New Year's Eve.
        self._add_new_years_eve(tr("Fiesta de Fin de Año"))


class VE(Venezuela):
    pass


class VEN(Venezuela):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, FEB, JUN, AUG, SEP, NOV, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class DominicanRepublic(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    http://ojd.org.do/Normativas/LABORAL/Leyes/Ley%20No.%20%20139-97.pdf
    https://es.wikipedia.org/wiki/Rep%C3%BAblica_Dominicana#D%C3%ADas_festivos_nacionales
    """

    country = "DO"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def _move_holiday(self, dt: date, include_sun: bool = False) -> None:
        # Law No. 139-97 - Holidays Dominican Republic - Jun 27, 1997
        if dt < date(1997, JUN, 27):
            return None
        obs_dt = None
        if self._is_tuesday(dt) or self._is_wednesday(dt):
            obs_dt = self._get_nth_weekday_from(-1, MON, dt)
        elif self._is_thursday(dt) or self._is_friday(dt) or (include_sun and self._is_sunday(dt)):
            obs_dt = self._get_nth_weekday_from(1, MON, dt)
        if obs_dt:
            self._add_holiday(self[dt], obs_dt)
            self.pop(dt)

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Epiphany.
        self._move_holiday(self._add_epiphany_day(tr("Día de los Santos Reyes")))

        # Lady of Altagracia.
        self._add_holiday(tr("Día de la Altagracia"), JAN, 21)

        # Juan Pablo Duarte Day.
        self._move_holiday(self._add_holiday(tr("Día de Duarte"), JAN, 26))

        # Independence Day.
        self._add_holiday(tr("Día de Independencia"), FEB, 27)

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Labor Day.
        self._move_holiday(self._add_labor_day(tr("Día del Trabajo")), include_sun=True)

        # Feast of Corpus Christi.
        self._add_corpus_christi_day(tr("Corpus Christi"))

        # Restoration Day.
        name = tr("Día de la Restauración")
        # Judgment No. 14 of Feb 20, 2008 of the Supreme Court of Justice
        if year <= 2007 and year % 4 == 0:
            self._add_holiday(name, AUG, 16)
        else:
            self._move_holiday(self._add_holiday(name, AUG, 16))

        # Our Lady of Mercedes Day.
        self._add_holiday(tr("Día de las Mercedes"), SEP, 24)

        # Constitution Day.
        self._move_holiday(self._add_holiday(tr("Día de la Constitución"), NOV, 6))

        # Christmas Day.
        self._add_christmas_day(tr("Día de Navidad"))


class DO(DominicanRepublic):
    pass


class DOM(DominicanRepublic):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import GREGORIAN_CALENDAR, AUG
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Moldova(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Moldova
    https://www.legis.md/cautare/getResults?doc_id=133686
    """

    country = "MD"
    default_language = "ro"
    supported_languages = ("en_US", "ro", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        if year <= 1990:
            return None
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Anul Nou"))

        name = (
            # Christmas (by old style).
            tr("Naşterea lui Iisus Hristos (Crăciunul pe stil vechi)")
            if year >= 2014
            # Christmas.
            else tr("Naşterea lui Iisus Hristos (Crăciunul)")
        )
        self._add_christmas_day(name)
        self._add_christmas_day_two(name)

        # International Women's Day.
        self._add_womens_day(tr("Ziua internatională a femeii"))

        # Easter.
        name = tr("Paştele")
        self._add_easter_sunday(name)
        self._add_easter_monday(name)

        # Day of Rejoicing.
        self._add_holiday(tr("Paştele blajinilor"), self._easter_sunday + td(days=+8))

        # International Workers' Solidarity Day.
        self._add_labor_day(tr("Ziua internaţională a solidarităţii oamenilor muncii"))

        may_9 = self._add_world_war_two_victory_day(
            # Victory Day and Commemoration of the heroes fallen for
            # Independence of Fatherland.
            tr("Ziua Victoriei şi a comemorării eroilor căzuţi pentru Independenţa Patriei")
        )

        if year >= 2017:
            # Europe Day.
            self._add_holiday(tr("Ziua Europei"), may_9)

        if year >= 2016:
            # International Children's Day
            self._add_childrens_day(tr("Ziua Ocrotirii Copilului"))

        # Republic of Moldova Independence Day
        self._add_holiday(tr("Ziua independenţei Republicii Moldova"), AUG, 27)

        # National Language Day.
        self._add_holiday(tr("Limba noastră"), AUG, 31)

        if year >= 2013:
            self._add_christmas_day(
                # Christmas (by new style).
                tr("Naşterea lui Iisus Hristos (Crăciunul pe stil nou)"),
                GREGORIAN_CALENDAR,
            )


class MD(Moldova):
    pass


class MDA(Moldova):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, MAY, JUL, AUG, SEP, OCT, NOV
from holidays.constants import PUBLIC, WORKDAY
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Slovakia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Slovakia
    https://sk.wikipedia.org/wiki/Zoznam_sviatkov_na_Slovensku
    https://www.slov-lex.sk/pravne-predpisy/SK/ZZ/1993/241/
    """

    country = "SK"
    default_language = "sk"
    special_public_holidays = {
        # 100th anniversary of the adoption of the Declaration
        # of the Slovak Nation.
        2018: (OCT, 30, tr("100. výročie prijatia Deklarácie slovenského národa"))
    }
    supported_categories = {PUBLIC, WORKDAY}
    supported_languages = ("en_US", "sk", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # Independent Slovak Republic established on Jan 01, 1993
        if self._year <= 1992:
            return None

        # Day of the Establishment of the Slovak Republic.
        self._add_holiday(tr("Deň vzniku Slovenskej republiky"), JAN, 1)

        # Epiphany.
        self._add_epiphany_day(
            tr("Zjavenie Pána (Traja králi a vianočnýsviatok pravoslávnych kresťanov)")
        )

        # Good Friday.
        self._add_good_friday(tr("Veľký piatok"))

        # Easter Monday.
        self._add_easter_monday(tr("Veľkonočný pondelok"))

        # Labor Day.
        self._add_labor_day(tr("Sviatok práce"))

        # Day of Victory over Fascism.
        if self._year >= 1997:
            # Day of Victory over Fascism.
            self._add_holiday(tr("Deň víťazstva nad fašizmom"), MAY, 8)

        # St. Cyril and Methodius Day.
        self._add_holiday(tr("Sviatok svätého Cyrila a svätého Metoda"), JUL, 5)

        # Slovak National Uprising Anniversary.
        self._add_holiday(tr("Výročie Slovenského národného povstania"), AUG, 29)

        # Constitution Day.
        self._add_holiday(tr("Deň Ústavy Slovenskej republiky"), SEP, 1)

        # Day of Our Lady of the Seven Sorrows.
        self._add_holiday(tr("Sedembolestná Panna Mária"), SEP, 15)

        # All Saints' Day.
        self._add_all_saints_day(tr("Sviatok Všetkých svätých"))

        if self._year >= 2001:
            # Struggle for Freedom and Democracy Day.
            self._add_holiday(tr("Deň boja za slobodu a demokraciu"), NOV, 17)

        # Christmas Eve.
        self._add_christmas_eve(tr("Štedrý deň"))

        # Christmas Day.
        self._add_christmas_day(tr("Prvý sviatok vianočný"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Druhý sviatok vianočný"))

    def _populate_workday_holidays(self):
        # According to Law 241/1993, this state holiday is not a non-working day.
        if self._year >= 2021:
            # Day of the Establishment of the Independent Czech-Slovak State.
            self._add_holiday(tr("Deň vzniku samostatného česko-slovenského štátu"), OCT, 28)


class SK(Slovakia):
    pass


class SVK(Slovakia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import APR, SEP, OCT, WED
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Honduras(HolidayBase, ChristianHolidays, InternationalHolidays):
    # Artículo 339 del Código del Trabajo:
    # https://www.ilo.org/dyn/natlex/docs/WEBTEXT/29076/64849/S59HND01.htm

    country = "HN"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Maundy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Holy Saturday.
        self._add_holy_saturday(tr("Sábado de Gloria"))

        # Panamerican Day.
        self._add_holiday(tr("Día de las Américas"), APR, 14)

        # Labor Day.
        self._add_labor_day(tr("Día del Trabajo"))

        # Independence Day.
        self._add_holiday(tr("Día de la Independencia"), SEP, 15)

        # https://www.tsc.gob.hn/web/leyes/Decreto_78-2015_Traslado_de_Feriados_Octubre.pdf
        if year <= 2014:
            # Morazan's Day.
            self._add_holiday(tr("Día de Morazán"), OCT, 3)

            # Columbus Day.
            self._add_columbus_day(tr("Día de la Raza"))

            # Army Day.
            self._add_holiday(tr("Día de las Fuerzas Armadas"), OCT, 21)
        else:
            # Morazan Weekend.
            holiday_name = tr("Semana Morazánica")
            # First Wednesday of October from 12 noon to Saturday 12 noon.
            first_wednesday = self._get_nth_weekday_of_month(1, WED, OCT)
            self._add_holiday(holiday_name, first_wednesday)
            self._add_holiday(holiday_name, first_wednesday + td(days=+1))
            self._add_holiday(holiday_name, first_wednesday + td(days=+2))

        # Christmas.
        self._add_christmas_day(tr("Navidad"))


class HN(Honduras):
    pass


class HND(Honduras):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)
#  Copyright: Kateryna Golovanova <kate@kgthreads.com>, 2022

from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, APR, MAY, JUN, JUL, AUG, SEP, NOV
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Bolivia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Bolivia
    https://www.officeholidays.com/countries/bolivia
    """

    country = "BO"
    subdivisions = (
        "B",
        "C",
        "H",
        "L",
        "N",
        "O",
        "P",
        "S",
        "T",
    )

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)
        observed_dates = set()

        # New Year's Day.
        if year >= 1825:
            observed_dates.add(self._add_new_years_day("Año Nuevo"))

        # Plurinational State Foundation Day.
        if year >= 2010:
            self._add_holiday("Nacimiento del Estado Plurinacional de Bolivia", JAN, 22)

        # Carnival.
        name = "Feriado por Carnaval"
        self._add_carnival_monday(name)
        self._add_carnival_tuesday(name)

        # Good Friday.
        self._add_good_friday("Viernes Santo")

        # Labor Day.
        observed_dates.add(self._add_labor_day("Día del trabajo"))

        # Corpus Christi.
        self._add_corpus_christi_day("Corpus Christi")

        if year >= 2010:
            # Andean New Year.
            observed_dates.add(self._add_holiday("Año Nuevo Andino", JUN, 21))

        if year >= 1825:
            # Independence Day.
            observed_dates.add(self._add_holiday("Día de la Patria", AUG, 6))

        # All Soul's Day.
        observed_dates.add(self._add_all_souls_day("Todos Santos"))

        # Christmas Day.
        observed_dates.add(self._add_christmas_day("Navidad"))

        if self.observed:
            for dt in sorted(observed_dates):
                if not self._is_sunday(dt):
                    continue
                self._add_holiday(f"{self[dt]} (Observed)", dt + td(days=+1))

    def _add_subdiv_b_holidays(self):
        # Beni Day.
        self._add_holiday("Día del departamento de Beni", NOV, 18)

    def _add_subdiv_c_holidays(self):
        # Cochabamba Day.
        self._add_holiday("Día del departamento de Cochabamba", SEP, 14)

    def _add_subdiv_h_holidays(self):
        # Chuquisaca Day.
        self._add_holiday("Día del departamento de Chuquisaca", MAY, 25)

    def _add_subdiv_l_holidays(self):
        # La Paz Day.
        self._add_holiday("Día del departamento de La Paz", JUL, 16)

    def _add_subdiv_n_holidays(self):
        # Pando Day.
        self._add_holiday("Día del departamento de Pando", SEP, 24)

    def _add_subdiv_p_holidays(self):
        # Potosí Day.
        self._add_holiday("Día del departamento de Potosí", NOV, 10)

    def _add_subdiv_o_holidays(self):
        # Carnival in Oruro.
        self._add_holiday("Carnaval de Oruro", self._easter_sunday + td(days=-51))

    def _add_subdiv_s_holidays(self):
        # Santa Cruz Day.
        self._add_holiday("Día del departamento de Santa Cruz", SEP, 24)

    def _add_subdiv_t_holidays(self):
        # La Tablada.
        self._add_holiday("La Tablada", APR, 15)


class BO(Bolivia):
    pass


class BOL(Bolivia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars import _CustomIslamicCalendar
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays, IslamicHolidays


class Cameroon(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Cameroon
      - https://www.timeanddate.com/holidays/cameroon
      - https://www.officeholidays.com/countries/cameroon
    """

    country = "CM"
    special_holidays = {
        2021: (
            (MAY, 14, "Public Holiday"),
            (JUL, 19, "Public Holiday"),
        ),
    }

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=CameroonIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # On 1 January 1960, French Cameroun gained independence from France.
        if year <= 1959:
            return None

        super()._populate(year)
        observed_dates = set()

        # New Year's Day.
        observed_dates.add(self._add_new_years_day("New Year's Day"))

        # Youth Day.
        if year >= 1966:
            observed_dates.add(self._add_holiday("Youth Day", FEB, 11))

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Labour Day.
        observed_dates.add(self._add_labor_day("Labour Day"))

        # National Day.
        if year >= 1972:
            observed_dates.add(self._add_holiday("National Day", MAY, 20))

        # Ascension Day.
        self._add_ascension_thursday("Ascension Day")

        # Assumption Day.
        observed_dates.add(self._add_assumption_of_mary_day("Assumption Day"))

        # Christmas Day.
        observed_dates.add(self._add_christmas_day("Christmas Day"))

        # Eid al-Fitr.
        observed_dates.update(self._add_eid_al_fitr_day("Eid al-Fitr"))

        # Eid al-Adha.
        observed_dates.update(self._add_eid_al_adha_day("Eid al-Adha"))

        # Mawlid.
        observed_dates.update(self._add_mawlid_day("Mawlid"))

        if self.observed:
            for dt in sorted(observed_dates):
                if not self._is_sunday(dt):
                    continue
                obs_date = dt + td(days=+1)
                while obs_date in observed_dates:
                    obs_date += td(days=+1)
                for hol_name in self.get_list(dt):
                    observed_dates.add(self._add_holiday("%s (Observed)" % hol_name, obs_date))

            # observed holidays special cases
            if year == 2007:
                self._add_holiday("Eid al-Adha (Observed)", JAN, 2)


class CM(Cameroon):
    pass


class CMR(Cameroon):
    pass


class CameroonIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2004: (FEB, 2),
        2005: (JAN, 21),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 9),
        2009: (NOV, 28),
        2010: (NOV, 17),
        2011: (NOV, 7),
        2012: (OCT, 26),
        2013: (OCT, 15),
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 13),
        2017: (SEP, 2),
        2018: (AUG, 21),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 9),
        2023: (JUN, 28),
    }

    EID_AL_FITR_DATES = {
        2001: (DEC, 17),
        2002: (DEC, 6),
        2003: (NOV, 26),
        2004: (NOV, 14),
        2005: (NOV, 4),
        2006: (OCT, 24),
        2007: (OCT, 13),
        2008: (OCT, 2),
        2009: (SEP, 21),
        2010: (SEP, 10),
        2011: (AUG, 31),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 28),
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2018: (JUN, 15),
        2019: (JUN, 4),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 21),
    }

    MAWLID_DATES = {
        2001: (JUN, 4),
        2002: (MAY, 24),
        2003: (MAY, 14),
        2004: (MAY, 2),
        2005: (APR, 21),
        2006: (APR, 11),
        2007: (MAR, 31),
        2008: (MAR, 20),
        2009: (MAR, 9),
        2010: (FEB, 26),
        2011: (FEB, 16),
        2012: (FEB, 5),
        2013: (JAN, 24),
        2014: (JAN, 14),
        2015: ((JAN, 3), (DEC, 24)),
        2016: (DEC, 12),
        2017: (DEC, 1),
        2018: (NOV, 21),
        2019: (NOV, 10),
        2020: (OCT, 29),
        2021: (OCT, 19),
        2022: (OCT, 8),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import MAY, AUG, OCT, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Jamaica(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Jamaica
    https://www.mlss.gov.jm/wp-content/uploads/2017/11/The-Holidays-Public-General-Act.pdf
    """

    country = "JM"

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, include_sat: bool = False, days: int = +1) -> None:
        if not self.observed:
            return None
        if self._is_sunday(dt) or (include_sat and self._is_saturday(dt)):
            self._add_holiday(
                "%s (Observed)" % self[dt], dt + td(days=+2 if self._is_saturday(dt) else days)
            )

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self._add_observed(self._add_new_years_day("New Year's Day"))

        # Ash Wednesday
        self._add_ash_wednesday("Ash Wednesday")

        # Good Friday
        self._add_good_friday("Good Friday")

        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # National Labour Day
        self._add_observed(self._add_holiday("National Labour Day", MAY, 23), include_sat=True)

        # Emancipation Day
        if year >= 1998:
            self._add_observed(self._add_holiday("Emancipation Day", AUG, 1))

        # Independence Day
        self._add_observed(self._add_holiday("Independence Day", AUG, 6))

        # National Heroes Day
        self._add_holiday("National Heroes Day", self._get_nth_weekday_of_month(3, MON, OCT))

        # Christmas Day
        self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)

        # Boxing Day
        self._add_observed(self._add_christmas_day_two("Boxing Day"))


class JM(Jamaica):
    pass


class JAM(Jamaica):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import FEB, MAY, JUN, AUG
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Estonia(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "EE"
    default_language = "et"
    supported_languages = ("en_US", "et", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("uusaasta"))

        # Independence Day.
        self._add_holiday(tr("iseseisvuspäev"), FEB, 24)

        # Good Friday.
        self._add_good_friday(tr("suur reede"))

        # Easter Sunday.
        self._add_easter_sunday(tr("ülestõusmispühade 1. püha"))

        # Spring Day.
        self._add_holiday(tr("kevadpüha"), MAY, 1)

        # Whit Sunday.
        self._add_whit_sunday(tr("nelipühade 1. püha"))

        # Victory Day.
        self._add_holiday(tr("võidupüha"), JUN, 23)

        # Midsummer Day.
        self._add_saint_johns_day(tr("jaanipäev"))

        if year >= 1998:
            # Independence Restoration Day.
            self._add_holiday(tr("taasiseseisvumispäev"), AUG, 20)

        if year >= 2005:
            # Christmas Eve.
            self._add_christmas_eve(tr("jõululaupäev"))

        # Christmas Day.
        self._add_christmas_day(tr("esimene jõulupüha"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("teine jõulupüha"))


class EE(Estonia):
    pass


class EST(Estonia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Paraguay(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.ghp.com.py/news/feriados-nacionales-del-ano-2019-en-paraguay
    https://es.wikipedia.org/wiki/Anexo:D%C3%ADas_feriados_en_Paraguay
    http://www.calendarioparaguay.com/
    """

    country = "PY"
    default_language = "es"

    # Public holiday.
    public_holiday = tr("Asueto adicionale")
    # Public sector holiday.
    public_sector_holiday = tr("Asueto de la Administración Pública")
    special_holidays = {
        # public holiday for business purposes, in view of
        # the recently increased risk of Dengue fever
        2007: (JAN, 29, public_holiday),
        # public sector holiday to celebrate Paraguay's
        # football team's qualification for the 2010 World Cup
        2009: (SEP, 10, public_holiday),
        2010: (
            # public holiday to coincide with the Paraguay-Italy
            # game of the current World Cup
            (JUN, 14, public_holiday),
            # 2 year-end public sector holidays
            (DEC, 24, public_sector_holiday),
            (DEC, 31, public_sector_holiday),
        ),
        2011: (
            # public holiday to coincide with the current anti-Dengue drive
            (APR, 19, public_holiday),
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 20, public_sector_holiday),
            # public holidays to commemorate the Bicentennial
            # of Paraguay's independence
            (MAY, 14, public_holiday),
            (MAY, 16, public_holiday),
            # 2 year-end public sector holidays
            (DEC, 23, public_sector_holiday),
            (DEC, 30, public_sector_holiday),
        ),
        2012: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 4, public_sector_holiday),
            # 2 year-end public sector holidays
            (DEC, 24, public_sector_holiday),
            (DEC, 31, public_sector_holiday),
        ),
        2013: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (MAR, 27, public_sector_holiday),
            # date of the inauguration of President-elect
            # Horacio Cartes, as a one-off non-working public holiday
            (AUG, 14, public_holiday),
        ),
        2014: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 16, public_sector_holiday),
            # 2 year-end public sector holidays
            (DEC, 24, public_sector_holiday),
            (DEC, 31, public_sector_holiday),
        ),
        2015: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 1, public_sector_holiday),
            # public holidays in Paraguay on account
            # of the upcoming visit of Pope Francis in Paraguay
            (JUL, 10, public_holiday),
            (JUL, 11, public_holiday),
            # 2 year-end public sector holidays
            (DEC, 24, public_sector_holiday),
            (DEC, 31, public_sector_holiday),
        ),
        # public sector holiday to let civil servants
        # begin their Holy Week earlier
        2016: (MAR, 23, public_sector_holiday),
        # public sector holiday to let civil servants
        # begin their Holy Week earlier
        2017: (MAR, 28, public_sector_holiday),
        2018: (
            # 2 year-end public sector holidays
            (DEC, 24, public_sector_holiday),
            (DEC, 31, public_sector_holiday),
        ),
        2019: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 17, public_sector_holiday),
            # 2 year-end public sector holidays
            (DEC, 24, public_sector_holiday),
            (DEC, 31, public_sector_holiday),
        ),
        # public sector holiday to let civil servants
        # begin their Holy Week earlier
        2020: (APR, 8, public_sector_holiday),
        2021: (
            # 2 year-end public sector holidays
            (DEC, 24, public_sector_holiday),
            (DEC, 31, public_sector_holiday),
        ),
        2022: (
            # public sector holiday to let civil servants
            # begin their Holy Week earlier
            (APR, 13, public_sector_holiday),
            # public sector holiday due to the annual May 1st
            # public holiday falling on a Sunday
            (MAY, 2, public_sector_holiday),
        ),
    }
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _move_holiday(self, dt: date) -> None:
        if not self.observed and self._is_weekend(dt):
            self.pop(dt)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._move_holiday(self._add_new_years_day(tr("Año Nuevo")))

        dates_obs = {
            2013: (MAR, 4),
            2016: (FEB, 29),
            2018: (FEB, 26),
            2022: (FEB, 28),
        }
        # Patriots Day.
        self._add_holiday(tr("Día de los Héroes de la Patria"), *dates_obs.get(year, (MAR, 1)))

        # Maundy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Easter Day.
        self._move_holiday(self._add_easter_sunday(tr("Día de Pascuas")))

        # Labour Day.
        self._move_holiday(self._add_labor_day(tr("Día del Trabajador")))

        # Independence Day.
        name = tr("Día de la Independencia Nacional")
        if year >= 2012:
            self._move_holiday(self._add_holiday(name, MAY, 14))
        may_15 = self._add_holiday(name, MAY, 15)
        if year != 2021:
            self._move_holiday(may_15)

        dates_obs = {
            2014: (JUN, 16),
            2018: (JUN, 11),
        }
        self._move_holiday(
            # Chaco Armistice Day.
            self._add_holiday(tr("Día de la Paz del Chaco"), *dates_obs.get(year, (JUN, 12)))
        )

        # Asuncion Foundation's Day.
        self._move_holiday(self._add_holiday(tr("Día de la Fundación de Asunción"), AUG, 15))

        if year >= 2000:
            dates_obs = {
                2015: (SEP, 28),
                2016: (OCT, 3),
                2017: (OCT, 2),
                2021: (SEP, 27),
                2022: (OCT, 3),
            }
            self._move_holiday(
                self._add_holiday(
                    # Boqueron Battle Day.
                    tr("Día de la Batalla de Boquerón"),
                    *dates_obs.get(year, (SEP, 29)),
                )
            )

        # Caacupe Virgin Day.
        self._move_holiday(self._add_holiday(tr("Día de la Virgen de Caacupé"), DEC, 8))

        # Christmas.
        self._add_christmas_day(tr("Navidad"))


class PY(Paraguay):
    pass


class PRY(Paraguay):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

__all__ = (
    "country_holidays",
    "CountryHoliday",
    "financial_holidays",
    "list_localized_countries",
    "list_localized_financial",
    "list_supported_countries",
    "list_supported_financial",
)

import warnings
from functools import lru_cache
from typing import Dict, Iterable, List, Optional, Tuple, Union

from holidays.holiday_base import HolidayBase
from holidays.registry import EntityLoader


def country_holidays(
    country: str,
    subdiv: Optional[str] = None,
    years: Optional[Union[int, Iterable[int]]] = None,
    expand: bool = True,
    observed: bool = True,
    prov: Optional[str] = None,
    state: Optional[str] = None,
    language: Optional[str] = None,
    categories: Optional[Tuple[str]] = None,
) -> HolidayBase:
    """
    Returns a new dictionary-like :py:class:`HolidayBase` object for the public
    holidays of the country matching **country** and other keyword arguments.

    :param country:
        An ISO 3166-1 Alpha-2 country code.

    :param subdiv:
        The subdivision (e.g. state or province); not implemented for all
        countries (see documentation).

    :param years:
        The year(s) to pre-calculate public holidays for at instantiation.

    :param expand:
        Whether the entire year is calculated when one date from that year
        is requested.

    :param observed:
        Whether to include the dates of when public holiday are observed
        (e.g. a holiday falling on a Sunday being observed the following
        Monday). False may not work for all countries.

    :param prov:
        *deprecated* use subdiv instead.

    :param state:
        *deprecated* use subdiv instead.

    :param language:
        The language which the returned holiday names will be translated
        into. It must be an ISO 639-1 (2-letter) language code. If the
        language translation is not supported the original holiday names
        will be used.

    :param categories:
        Requested holiday categories.

    :return:
        A :py:class:`HolidayBase` object matching the **country**.

    The key of the :class:`dict`-like :class:`HolidayBase` object is the
    `date` of the holiday, and the value is the name of the holiday itself.
    Dates where a key is not present are not public holidays (or, if
    **observed** is False, days when a public holiday is observed).

    When passing the `date` as a key, the `date` can be expressed in one of the
    following types:

    * :class:`datetime.date`,
    * :class:`datetime.datetime`,
    * a :class:`str` of any format recognized by :func:`dateutil.parser.parse`,
    * or a :class:`float` or :class:`int` representing a POSIX timestamp.

    The key is always returned as a :class:`datetime.date` object.

    To maximize speed, the list of public holidays is built on the fly as
    needed, one calendar year at a time. When the object is instantiated
    without a **years** parameter, it is empty, but, unless **expand** is set
    to False, as soon as a key is accessed the class will calculate that entire
    year's list of holidays and set the keys with them.

    If you need to list the holidays as opposed to querying individual dates,
    instantiate the class with the **years** parameter.

    Example usage:

    >>> from holidays import country_holidays
    >>> us_holidays = country_holidays('US')
    # For a specific subdivision (e.g. state or province):
    >>> calif_holidays = country_holidays('US', subdiv='CA')

    The below will cause 2015 holidays to be calculated on the fly:

    >>> from datetime import date
    >>> assert date(2015, 1, 1) in us_holidays

    This will be faster because 2015 holidays are already calculated:

    >>> assert date(2015, 1, 2) not in us_holidays

    The :class:`HolidayBase` class also recognizes strings of many formats
    and numbers representing a POSIX timestamp:

    >>> assert '2014-01-01' in us_holidays
    >>> assert '1/1/2014' in us_holidays
    >>> assert 1388597445 in us_holidays

    Show the holiday's name:

    >>> us_holidays.get('2014-01-01')
    "New Year's Day"

    Check a range:

    >>> us_holidays['2014-01-01': '2014-01-03']
    [datetime.date(2014, 1, 1)]

    List all 2020 holidays:

    >>> us_holidays = country_holidays('US', years=2020)
    >>> for day in us_holidays.items():
    ...     print(day)
    (datetime.date(2020, 1, 1), "New Year's Day")
    (datetime.date(2020, 1, 20), 'Martin Luther King Jr. Day')
    (datetime.date(2020, 2, 17), "Washington's Birthday")
    (datetime.date(2020, 5, 25), 'Memorial Day')
    (datetime.date(2020, 7, 4), 'Independence Day')
    (datetime.date(2020, 7, 3), 'Independence Day (Observed)')
    (datetime.date(2020, 9, 7), 'Labor Day')
    (datetime.date(2020, 10, 12), 'Columbus Day')
    (datetime.date(2020, 11, 11), 'Veterans Day')
    (datetime.date(2020, 11, 26), 'Thanksgiving')
    (datetime.date(2020, 12, 25), 'Christmas Day')

    Some holidays are only present in parts of a country:

    >>> us_pr_holidays = country_holidays('US', subdiv='PR')
    >>> assert '2018-01-06' not in us_holidays
    >>> assert '2018-01-06' in us_pr_holidays

    Append custom holiday dates by passing one of:

    * a :class:`dict` with date/name key/value pairs (e.g.
      ``{'2010-07-10': 'My birthday!'}``),
    * a list of dates (as a :class:`datetime.date`, :class:`datetime.datetime`,
      :class:`str`, :class:`int`, or :class:`float`); ``'Holiday'`` will be
      used as a description,
    * or a single date item (of one of the types above); ``'Holiday'`` will be
      used as a description:

    >>> custom_holidays = country_holidays('US', years=2015)
    >>> custom_holidays.update({'2015-01-01': "New Year's Day"})
    >>> custom_holidays.update(['2015-07-01', '07/04/2015'])
    >>> custom_holidays.update(date(2015, 12, 25))
    >>> assert date(2015, 1, 1) in custom_holidays
    >>> assert date(2015, 1, 2) not in custom_holidays
    >>> assert '12/25/2015' in custom_holidays

    For more complex logic, like 4th Monday of January, you can inherit the
    :class:`HolidayBase` class and define your own :meth:`_populate` method.
    See documentation for examples.
    """
    import holidays

    try:
        return getattr(holidays, country)(
            years=years,
            subdiv=subdiv,
            expand=expand,
            observed=observed,
            prov=prov,
            state=state,
            language=language,
            categories=categories,
        )
    except AttributeError:
        raise NotImplementedError(f"Country {country} not available")


def financial_holidays(
    market: str,
    subdiv: Optional[str] = None,
    years: Optional[Union[int, Iterable[int]]] = None,
    expand: bool = True,
    observed: bool = True,
    language: Optional[str] = None,
) -> HolidayBase:
    """
    Returns a new dictionary-like :py:class:`HolidayBase` object for the public
    holidays of the financial market matching **market** and other keyword
    arguments.

    :param market:
        An ISO 3166-1 Alpha-2 market code.

    :param subdiv:
        Currently not implemented for markets (see documentation).

    :param years:
        The year(s) to pre-calculate public holidays for at instantiation.

    :param expand:
        Whether the entire year is calculated when one date from that year
        is requested.

    :param observed:
        Whether to include the dates of when public holiday are observed
        (e.g. a holiday falling on a Sunday being observed the following
        Monday). False may not work for all countries.

    :param language:
        The language which the returned holiday names will be translated
        into. It must be an ISO 639-1 (2-letter) language code. If the
        language translation is not supported the original holiday names
        will be used.

    :return:
        A :py:class:`HolidayBase` object matching the **market**.

    Example usage:

    >>> from holidays import financial_holidays
    >>> nyse_holidays = financial_holidays('NYSE')

    See :py:func:`country_holidays` documentation for further details and
    examples.
    """
    import holidays

    try:
        return getattr(holidays, market)(
            years=years,
            subdiv=subdiv,
            expand=expand,
            observed=observed,
            language=language,
        )
    except AttributeError:
        raise NotImplementedError(f"Financial market {market} not available")


def CountryHoliday(
    country: str,
    subdiv: Optional[str] = None,
    years: Optional[Union[int, Iterable[int]]] = None,
    expand: bool = True,
    observed: bool = True,
    prov: Optional[str] = None,
    state: Optional[str] = None,
) -> HolidayBase:
    """
    Deprecated name for :py:func:`country_holidays`.

    :meta private:
    """

    warnings.warn(
        "CountryHoliday is deprecated, use country_holidays instead.", DeprecationWarning
    )
    return country_holidays(country, subdiv, years, expand, observed, prov, state)


def _list_localized_entities(entity_codes: Iterable[str]) -> Dict[str, List[str]]:
    """
    Get all localized entities and languages they support.

    :param entity_codes:
        A list of entity codes.

    :return:
        A dictionary where key is an entity code and
        value is a list of supported languages (either ISO 639-1 or a
        combination of ISO 639-1 and ISO 3166-1 codes joined with "_").
    """
    import holidays

    localized_countries = {}
    for entity_code in entity_codes:
        languages = getattr(holidays, entity_code).supported_languages
        if len(languages) == 0:
            continue
        localized_countries[entity_code] = sorted(languages)

    return localized_countries


@lru_cache()
def list_localized_countries(include_aliases=True) -> Dict[str, List[str]]:
    """
    Get all localized countries and languages they support.

    :param include_aliases:
        Whether to include entity aliases (e.g. UK for GB).

    :return:
        A dictionary where key is an ISO 3166-1 alpha-2 country code and
        value is a list of supported languages (either ISO 639-1 or a
        combination of ISO 639-1 and ISO 3166-1 codes joined with "_").
    """

    return _list_localized_entities(EntityLoader.get_country_codes(include_aliases))


@lru_cache()
def list_localized_financial(include_aliases=True) -> Dict[str, List[str]]:
    """
    Get all localized financial markets and languages they support.

    :param include_aliases:
        Whether to include entity aliases(e.g. TAR for ECB, XNYS for NYSE).

    :return:
        A dictionary where key is a market code and value is a list of
        supported subdivision codes.
    """

    return _list_localized_entities(EntityLoader.get_financial_codes(include_aliases))


def _list_supported_entities(entity_codes: Iterable[str]) -> Dict[str, List[str]]:
    """
    Get all supported entities and their subdivisions.

    :param entity_codes:
        A list of entity codes.

    :return:
        A dictionary where key is an entity code and value is a list
        of supported subdivision codes.
    """
    import holidays

    return {
        country_code: list(getattr(holidays, country_code).subdivisions)
        for country_code in entity_codes
    }


@lru_cache()
def list_supported_countries(include_aliases=True) -> Dict[str, List[str]]:
    """
    Get all supported countries and their subdivisions.

    :param include_aliases:
        Whether to include entity aliases (e.g. UK for GB).

    :return:
        A dictionary where key is an ISO 3166-1 alpha-2 country code and
        value is a list of supported subdivision codes.
    """
    return _list_supported_entities(EntityLoader.get_country_codes(include_aliases))


@lru_cache()
def list_supported_financial(include_aliases=True) -> Dict[str, List[str]]:
    """
    Get all supported financial markets and their subdivisions.

    :param include_aliases:
        Whether to include entity aliases(e.g. TAR for ECB, XNYS for NYSE).

    :return:
        A dictionary where key is a market code and value is a list of
        supported subdivision codes.
    """
    return _list_supported_entities(EntityLoader.get_financial_codes(include_aliases))



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, APR, MAY, JUN, JUL, AUG, OCT, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Uruguay(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Uruguay
    """

    country = "UY"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # Mandatory paid holidays:

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # International Workers' Day.
        self._add_labor_day(tr("Día de los Trabajadores"))

        # Constitution Day.
        self._add_holiday(tr("Jura de la Constitución"), JUL, 18)

        # Independence Day.
        self._add_holiday(tr("Día de la Independencia"), AUG, 25)

        # Day of the Family.
        self._add_christmas_day(tr("Día de la Familia"))

        # Partially paid holidays:

        # Children's Day.
        self._add_holiday(tr("Día de los Niños"), JAN, 6)

        # Birthday of José Gervasio Artigas.
        self._add_holiday(tr("Natalicio de José Gervasio Artigas"), JUN, 19)

        # All Souls' Day.
        self._add_all_souls_day(tr("Día de los Difuntos"))

        # Moveable holidays:

        # Carnival Day.
        name = tr("Día de Carnaval")
        self._add_carnival_monday(name)
        self._add_carnival_tuesday(name)

        # Maundy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))
        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))
        # Easter Day.
        self._add_easter_sunday(tr("Día de Pascuas"))

        holiday_pairs = (
            # Landing of the 33 Patriots.
            (date(year, APR, 19), tr("Desembarco de los 33 Orientales")),
            # Battle of Las Piedras.
            (date(year, MAY, 18), tr("Batalla de Las Piedras")),
            # Respect for Cultural Diversity Day.
            (date(year, OCT, 12), tr("Día del Respeto a la Diversidad Cultural")),
        )

        for dt, name in holiday_pairs:
            if self._is_tuesday(dt) or self._is_wednesday(dt):
                self._add_holiday(name, self._get_nth_weekday_from(-1, MON, dt))
            elif self._is_thursday(dt) or self._is_friday(dt):
                self._add_holiday(name, self._get_nth_weekday_from(1, MON, dt))
            else:
                self._add_holiday(name, dt)


class UY(Uruguay):
    pass


class URY(Uruguay):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr
from typing import Tuple

from holidays.calendars.gregorian import MAY, JUN, JUL, AUG, SEP, OCT, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Chile(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - https://www.feriados.cl
    - http://www.feriadoschilenos.cl/ (excellent history)
    - https://es.wikipedia.org/wiki/Anexo:D%C3%ADas_feriados_en_Chile
    - Law 2.977 (established official Chile holidays in its current form)
    - Law 20.983 (Day after New Year's Day, if it's a Sunday)
    - Law 19.668 (floating Monday holiday)
    - Law 19.668 (Corpus Christi)
    - Law 2.200, (Labour Day)
    - Law 18.018 (Labour Day renamed)
    - Law 16.840, Law 18.432 (Saint Peter and Saint Paul)
    - Law 20.148 (Day of Virgin of Carmen)
    - Law 18.026 (Day of National Liberation)
    - Law 19.588, Law 19.793 (Day of National Unity)
    - Law 20.983 (National Holiday Friday preceding Independence Day)
    - Law 20.215 (National Holiday Monday preceding Independence Day)
    - Law 20.215 (National Holiday Friday following Army Day)
    - Decree-law 636, Law 8.223
    - Law 3.810 (Columbus Day)
    - Law 20.299 (National Day of the Evangelical and Protestant Churches)
    - Law 20.663 (Región de Arica y Parinacota)
    - Law 20.678 (Región de Ñuble)
    """

    country = "CL"
    default_language = "es"
    special_holidays = {
        # National Holiday.
        2022: (SEP, 16, tr("Feriado nacional")),
    }
    subdivisions = (
        "AI",
        "AN",
        "AP",
        "AR",
        "AT",
        "BI",
        "CO",
        "LI",
        "LL",
        "LR",
        "MA",
        "ML",
        "NB",
        "RM",
        "TA",
        "VS",
    )
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _move_holiday(self, dt: date) -> None:
        if self._year <= 1999 or self._is_monday(dt) or self._is_weekend(dt):
            return None
        self._add_holiday(
            self[dt],
            self._get_nth_weekday_from(1, MON, dt)
            if self._is_friday(dt)
            else self._get_nth_weekday_from(-1, MON, dt),
        )
        self.pop(dt)

    def _populate(self, year):
        if year <= 1914:
            return None

        super()._populate(year)

        # New Year's Day.
        jan_1 = self._add_new_years_day(tr("Año Nuevo"))
        if year >= 2017 and self._is_sunday(jan_1):
            self._add_new_years_day_two(tr("Feriado nacional"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Holy Saturday.
        self._add_holy_saturday(tr("Sábado Santo"))

        if year <= 1967:
            # Ascension of Jesus.
            self._add_ascension_thursday(tr("Ascensión del Señor"))

        if year <= 1967 or 1987 <= year <= 2006:
            self._add_holiday(
                # Corpus Christi.
                tr("Corpus Christi"),
                self._easter_sunday + td(days=+60 if year <= 1999 else +57),
            )

        if year >= 1932:
            # Labour Day.
            self._add_labor_day(tr("Día Nacional del Trabajo"))

        # Naval Glories Day.
        self._add_holiday(tr("Día de las Glorias Navales"), MAY, 21)

        if year >= 2021:
            self._add_holiday(
                # National Day of Indigenous Peoples.
                tr("Día Nacional de los Pueblos Indígenas"),
                *((JUN, 21) if year == 2021 else self._summer_solstice_date),
            )

        if year <= 1967 or year >= 1986:
            # Saint Peter and Saint Paul.
            self._move_holiday(self._add_saints_peter_and_paul_day(tr("San Pedro y San Pablo")))

        if year >= 2007:
            # Day of Virgin of Carmen.
            self._add_holiday(tr("Virgen del Carmen"), JUL, 16)

        # Assumption of Mary.
        self._add_assumption_of_mary_day(tr("Asunción de la Virgen"))

        if 1981 <= year <= 1998:
            # Day of National Liberation.
            self._add_holiday(tr("Día de la Liberación Nacional"), SEP, 11)
        elif 1999 <= year <= 2001:
            self._add_holiday(
                # Day of National Unity.
                tr("Día de la Unidad Nacional"),
                self._get_nth_weekday_of_month(1, MON, SEP),
            )

        if year >= 2017 and self._is_saturday(SEP, 18):
            # National Holiday.
            self._add_holiday(tr("Fiestas Patrias"), SEP, 17)

        if year >= 2007 and self._is_tuesday(SEP, 18):
            self._add_holiday(tr("Fiestas Patrias"), SEP, 17)

        # Independence Day.
        self._add_holiday(tr("Día de la Independencia"), SEP, 18)

        # Army Day.
        self._add_holiday(tr("Día de las Glorias del Ejército"), SEP, 19)

        if year >= 2008 and self._is_thursday(SEP, 19):
            self._add_holiday(tr("Fiestas Patrias"), SEP, 20)

        if 1932 <= year <= 1944:
            self._add_holiday(tr("Fiestas Patrias"), SEP, 20)

        if year >= 1922 and year != 1973:
            name = (
                # Meeting of Two Worlds' Day.
                tr("Día del Encuentro de dos Mundos")
                if year >= 2000
                # Columbus Day.
                else tr("Día de la Raza")
            )
            self._move_holiday(self._add_columbus_day(name))

        if year >= 2008:
            dt = date(year, OCT, 31)
            # This holiday is moved to the preceding Friday if it falls on a Tuesday,
            # or to the following Friday if it falls on a Wednesday.
            if self._is_wednesday(dt):
                dt += td(days=+2)
            elif self._is_tuesday(dt):
                dt += td(days=-4)
            # National Day of the Evangelical and Protestant Churches.
            self._add_holiday(tr("Día Nacional de las Iglesias Evangélicas y Protestantes"), dt)

        # All Saints Day.
        self._add_all_saints_day(tr("Día de Todos los Santos"))

        # Immaculate Conception.
        self._add_immaculate_conception_day(tr("La Inmaculada Concepción"))

        if 1944 <= year <= 1988:
            # Christmas Eve.
            self._add_christmas_eve(tr("Víspera de Navidad"))

        # Christmas.
        self._add_christmas_day(tr("Navidad"))

    def _add_subdiv_ap_holidays(self):
        if self._year >= 2013:
            # Assault and Capture of Cape Arica.
            self._add_holiday(tr("Asalto y Toma del Morro de Arica"), JUN, 7)

    def _add_subdiv_nb_holidays(self):
        if self._year >= 2014:
            # Nativity of Bernardo O'Higgins (Chillán and Chillán Viejo communes)
            name = tr("Nacimiento del Prócer de la Independencia (Chillán y Chillán Viejo)")
            self._add_holiday(name, AUG, 20)

    @property
    def _summer_solstice_date(self) -> Tuple[int, int]:
        day = 20
        if (self._year % 4 > 1 and self._year <= 2046) or (
            self._year % 4 > 2 and self._year <= 2075
        ):
            day = 21
        return JUN, day


class CL(Chile):
    pass


class CHL(Chile):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import GREGORIAN_CALENDAR, JUL, NOV
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Belarus(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Belarus holidays.

    References:
     - http://president.gov.by/en/holidays_en/
     - http://www.belarus.by/en/about-belarus/national-holidays
    """

    country = "BY"
    default_language = "be"
    supported_languages = ("be", "en_US")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # The current set of holidays came into force in 1998.
        # http://laws.newsby.org/documents/ukazp/pos05/ukaz05806.htm
        if year <= 1998:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Новы год"))

        # Jan 2nd is the national holiday (New Year) from 2020.
        # http://president.gov.by/uploads/documents/2019/464uk.pdf
        if year >= 2020:
            self._add_new_years_day_two(tr("Новы год"))

        # Orthodox Christmas Day.
        self._add_christmas_day(tr("Нараджэнне Хрыстова (праваслаўнае Раство)"))

        # Women's Day.
        self._add_womens_day(tr("Дзень жанчын"))

        # Radunitsa (Day of Rejoicing).
        self._add_rejoicing_day(tr("Радаўніца"))

        # Labour Day.
        self._add_labor_day(tr("Свята працы"))

        # Victory Day.
        self._add_world_war_two_victory_day(tr("Дзень Перамогі"))

        # Independence Day.
        self._add_holiday(tr("Дзень Незалежнасці Рэспублікі Беларусь (Дзень Рэспублікі)"), JUL, 3)

        # October Revolution Day.
        self._add_holiday(tr("Дзень Кастрычніцкай рэвалюцыі"), NOV, 7)

        # Catholic Christmas Day.
        self._add_christmas_day(tr("Нараджэнне Хрыстова (каталіцкае Раство)"), GREGORIAN_CALENDAR)


class BY(Belarus):
    pass


class BLR(Belarus):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

import warnings
from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, APR, JUL, SEP, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Eswatini(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://swazilii.org/sz/legislation/act/1938/71
    https://www.officeholidays.com/countries/swaziland
    """

    country = "SZ"
    special_holidays = {
        # https://mg.co.za/article/1999-12-09-swaziland-declares-bank-holidays/
        1999: (DEC, 31, "Y2K changeover"),
        2000: (JAN, 3, "Y2K changeover"),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        # As of 2021/1/1, whenever a public holiday falls on a Sunday
        # it rolls over to the following Monday
        if self.observed and self._is_sunday(dt) and self._year >= 2021:
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=days))

    def _populate(self, year):
        # Observed since 1939
        if year <= 1938:
            return None

        super()._populate(year)

        self._add_observed(self._add_new_years_day("New Year's Day"))

        self._add_good_friday("Good Friday")

        self._add_easter_monday("Easter Monday")

        self._add_ascension_thursday("Ascension Day")

        if year >= 1987:
            apr_19 = self._add_holiday("King's Birthday", APR, 19)
            self._add_observed(apr_19, days=+2 if apr_19 == self._easter_sunday else +1)

        if year >= 1969:
            apr_25 = self._add_holiday("National Flag Day", APR, 25)
            self._add_observed(apr_25, days=+2 if apr_25 == self._easter_sunday else +1)

        self._add_observed(self._add_labor_day("Worker's Day"))

        if year >= 1983:
            self._add_observed(self._add_holiday("Birthday of Late King Sobhuza", JUL, 22))

        self._add_observed(self._add_holiday("Independence Day", SEP, 6))

        self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)

        self._add_observed(self._add_christmas_day_two("Boxing Day"))


class Swaziland(Eswatini):
    def __init__(self, *args, **kwargs) -> None:
        warnings.warn("Swaziland is deprecated, use Eswatini instead.", DeprecationWarning)

        super().__init__(*args, **kwargs)


class SZ(Eswatini):
    pass


class SZW(Eswatini):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

# flake8: noqa: F401

from .albania import Albania, AL, ALB
from .algeria import Algeria, DZ, DZA
from .american_samoa import AmericanSamoa, AS, ASM, HolidaysAS
from .andorra import Andorra, AD, AND
from .angola import Angola, AO, AGO
from .argentina import Argentina, AR, ARG
from .armenia import Armenia, AM, ARM
from .aruba import Aruba, AW, ABW
from .australia import Australia, AU, AUS
from .austria import Austria, AT, AUT
from .azerbaijan import Azerbaijan, AZ, AZE
from .bahrain import Bahrain, BH, BAH
from .bangladesh import Bangladesh, BD, BGD
from .belarus import Belarus, BY, BLR
from .belgium import Belgium, BE, BEL
from .belize import Belize, BZ, BLZ
from .bolivia import Bolivia, BO, BOL
from .bosnia_and_herzegovina import BosniaAndHerzegovina, BA, BIH
from .botswana import Botswana, BW, BWA
from .brazil import Brazil, BR, BRA
from .brunei import Brunei, BN, BRN
from .bulgaria import Bulgaria, BG, BLG
from .burkina_faso import BurkinaFaso, BF, BFA
from .burundi import Burundi, BI, BDI
from .cambodia import Cambodia, KH, KHM
from .cameroon import Cameroon, CM, CMR
from .canada import Canada, CA, CAN
from .chad import Chad, TD, TCD
from .chile import Chile, CL, CHL
from .china import China, CN, CHN
from .colombia import Colombia, CO, COL
from .costa_rica import CostaRica, CR, CRI
from .croatia import Croatia, HR, HRV
from .cuba import Cuba, CU, CUB
from .curacao import Curacao, CW, CUW
from .cyprus import Cyprus, CY, CYP
from .czechia import Czechia, CZ, CZE
from .denmark import Denmark, DK, DNK
from .djibouti import Djibouti, DJ, DJI
from .dominican_republic import DominicanRepublic, DO, DOM
from .ecuador import Ecuador, EC, ECU
from .egypt import Egypt, EG, EGY
from .el_salvador import ElSalvador, SV, SLV
from .estonia import Estonia, EE, EST
from .eswatini import Eswatini, SZ, SZW, Swaziland
from .ethiopia import Ethiopia, ET, ETH
from .finland import Finland, FI, FIN
from .france import France, FR, FRA
from .gabon import Gabon, GA, GAB
from .georgia import Georgia, GE, GEO
from .germany import Germany, DE, DEU
from .greece import Greece, GR, GRC
from .guam import Guam, GU, GUM, HolidaysGU
from .guatemala import Guatemala, GT, GUA
from .honduras import Honduras, HN, HND
from .hongkong import HongKong, HK, HKG
from .hungary import Hungary, HU, HUN
from .iceland import Iceland, IS, ISL
from .india import India, IN, IND
from .indonesia import Indonesia, ID, IDN
from .ireland import Ireland, IE, IRL
from .isle_of_man import IsleOfMan, IM, IMN
from .israel import Israel, IL, ISR
from .italy import Italy, IT, ITA
from .jamaica import Jamaica, JM, JAM
from .japan import Japan, JP, JPN
from .kazakhstan import Kazakhstan, KZ, KAZ
from .kenya import Kenya, KE, KEN
from .kyrgyzstan import Kyrgyzstan, KG, KGZ
from .latvia import Latvia, LV, LVA
from .lesotho import Lesotho, LS, LSO
from .liechtenstein import Liechtenstein, LI, LIE
from .lithuania import Lithuania, LT, LTU
from .luxembourg import Luxembourg, LU, LUX
from .madagascar import Madagascar, MG, MDG
from .malawi import Malawi, MW, MWI
from .malaysia import Malaysia, MY, MYS
from .malta import Malta, MT, MLT
from .marshall_islands import MarshallIslands, MH, MHL, HolidaysMH
from .mexico import Mexico, MX, MEX
from .moldova import Moldova, MD, MDA
from .monaco import Monaco, MC, MCO
from .montenegro import Montenegro, ME, MNE
from .morocco import Morocco, MA, MOR
from .mozambique import Mozambique, MZ, MOZ
from .namibia import Namibia, NA, NAM
from .netherlands import Netherlands, NL, NLD
from .new_zealand import NewZealand, NZ, NZL
from .nicaragua import Nicaragua, NI, NIC
from .nigeria import Nigeria, NG, NGA
from .north_macedonia import NorthMacedonia, MK, MKD
from .northern_mariana_islands import NorthernMarianaIslands, MP, MNP, HolidaysMP
from .norway import Norway, NO, NOR
from .pakistan import Pakistan, PK, PAK
from .panama import Panama, PA, PAN
from .paraguay import Paraguay, PY, PRY
from .peru import Peru, PE, PER
from .philippines import Philippines, PH, PHL
from .poland import Poland, PL, POL
from .portugal import Portugal, PT, PRT
from .puerto_rico import PuertoRico, PR, PRI, HolidaysPR
from .romania import Romania, RO, ROU
from .russia import Russia, RU, RUS
from .san_marino import SanMarino, SM, SMR
from .saudi_arabia import SaudiArabia, SA, SAU
from .serbia import Serbia, RS, SRB
from .singapore import Singapore, SG, SGP
from .slovakia import Slovakia, SK, SVK
from .slovenia import Slovenia, SI, SVN
from .south_africa import SouthAfrica, ZA, ZAF
from .south_korea import SouthKorea, KR, KOR, Korea
from .spain import Spain, ES, ESP
from .sweden import Sweden, SE, SWE
from .switzerland import Switzerland, CH, CHE
from .taiwan import Taiwan, TW, TWN
from .thailand import Thailand, TH, THA
from .tunisia import Tunisia, TN, TUN
from .turkey import Turkey, TR, TUR
from .ukraine import Ukraine, UA, UKR
from .united_arab_emirates import UnitedArabEmirates, AE, ARE
from .united_kingdom import UnitedKingdom, GB, GBR, UK
from .united_states import UnitedStates, US, USA
from .united_states_minor_outlying_islands import (
    UnitedStatesMinorOutlyingIslands,
    UM,
    UMI,
    HolidaysUM,
)
from .united_states_virgin_islands import UnitedStatesVirginIslands, VI, VIR, HolidaysVI
from .uruguay import Uruguay, UY, URY
from .uzbekistan import Uzbekistan, UZ, UZB
from .vatican_city import VaticanCity, VA, VAT
from .venezuela import Venezuela, VE, VEN
from .vietnam import Vietnam, VN, VNM
from .zambia import Zambia, ZM, ZMB
from .zimbabwe import Zimbabwe, ZW, ZWE



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, MAR, SEP, OCT, NOV, DEC
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class Albania(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Albania
    """

    country = "AL"
    special_holidays = {
        2022: (MAR, 21, "Public Holiday"),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)
        observed_dates = set()

        # New Year's Day.
        name = "New Year's Day"
        observed_dates.add(self._add_new_years_day(name))
        observed_dates.add(self._add_new_years_day_two(name))

        # Summer Day.
        if year >= 2004:
            observed_dates.add(self._add_holiday("Summer Day", MAR, 14))

        # Nevruz.
        if year >= 1996:
            observed_dates.add(self._add_holiday("Nevruz", MAR, 22))

        # Easter.
        observed_dates.add(self._add_easter_sunday("Catholic Easter"))
        observed_dates.add(self._add_easter_sunday("Orthodox Easter", JULIAN_CALENDAR))

        # May Day.
        observed_dates.add(self._add_labor_day("May Day"))

        # Mother Teresa Day.
        if 2004 <= year <= 2017:
            observed_dates.add(self._add_holiday("Mother Teresa Beatification Day", OCT, 19))
        elif year >= 2018:
            observed_dates.add(self._add_holiday("Mother Teresa Canonization Day", SEP, 5))

        # Independence Day.
        observed_dates.add(self._add_holiday("Independence Day", NOV, 28))

        # Liberation Day.
        observed_dates.add(self._add_holiday("Liberation Day", NOV, 29))

        # National Youth Day.
        if year >= 2009:
            observed_dates.add(self._add_holiday("National Youth Day", DEC, 8))

        # Christmas Day.
        observed_dates.add(self._add_christmas_day("Christmas Day"))

        # Eid al-Fitr.
        observed_dates.update(self._add_eid_al_fitr_day("Eid al-Fitr"))

        # Eid al-Adha.
        observed_dates.update(self._add_eid_al_adha_day("Eid al-Adha"))

        if self.observed:
            for hol_date in sorted(observed_dates):
                if not self._is_weekend(hol_date):
                    continue
                obs_date = hol_date + td(days=+1)
                while self._is_weekend(obs_date) or obs_date in observed_dates:
                    obs_date += td(days=+1)
                for hol_name in self.get_list(hol_date):
                    observed_dates.add(self._add_holiday("%s (Observed)" % hol_name, obs_date))

            # observed holidays special cases
            if year == 2007:
                self._add_holiday("Eid al-Adha (Observed)", JAN, 3)


class AL(Albania):
    pass


class ALB(Albania):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from typing import Optional

from holidays.calendars.gregorian import FEB, APR, JUN, JUL, OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class Burundi(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    Burundian holidays
    Note that holidays falling on a sunday maybe observed
    on the following Monday.
    This depends on formal annoucemnts by the government,
    which only happens close to the date of the holiday.

    Primary sources:
    https://www.officeholidays.com/countries/burundi
    """

    country = "BI"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_holiday(self, *args) -> Optional[date]:
        dt = super()._add_holiday(*args)
        if self.observed and dt and self._is_sunday(dt):
            super()._add_holiday(
                "%s (Observed)" % self[dt], dt + td(days=+1)  # type: ignore[index, operator]
            )
        return dt

    def _populate(self, year):
        if year <= 1961:
            return None

        super()._populate(year)

        # New Year's Day
        self._add_new_years_day("New Year's Day")

        # Unity Day
        if year >= 1992:
            self._add_holiday("Unity Day", FEB, 5)

        # President Ntaryamira Day
        if year >= 1995:
            self._add_holiday("President Ntaryamira Day", APR, 6)

        # Labour Day
        self._add_labor_day("Labour Day")

        # Ascension Day
        self._add_ascension_thursday("Ascension Day")

        # President Nkurunziza Day
        if year >= 2022:
            self._add_holiday("President Nkurunziza Day", JUN, 8)

        # Independence Day
        self._add_holiday("Independence Day", JUL, 1)

        # Assumption Day
        self._add_assumption_of_mary_day("Assumption Day")

        # Prince Louis Rwagasore Day
        self._add_holiday("Prince Louis Rwagasore Day", OCT, 13)

        # President Ndadaye's Day
        if year >= 1994:
            self._add_holiday("President Ndadaye's Day", OCT, 21)

        # All Saints' Day
        self._add_all_saints_day("All Saints' Day")

        # Christmas Day
        self._add_christmas_day("Christmas Day")

        # Eid ul Fitr
        # date of observance is announced yearly
        self._add_eid_al_fitr_day("Eid ul Fitr")

        # Eid al Adha
        # date of observance is announced yearly
        self._add_eid_al_adha_day("Eid al Adha")


class BI(Burundi):
    pass


class BDI(Burundi):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


from gettext import gettext as tr

from holidays.calendars.gregorian import GREGORIAN_CALENDAR, MAR, APR, OCT
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Cyprus(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Cyprus holidays.

    References:
     - https://en.wikipedia.org/wiki/Public_holidays_in_Cyprus
    """

    country = "CY"
    default_language = "el"
    supported_languages = ("el", "en_US")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Years Day.
        self._add_new_years_day(tr("Πρωτοχρονιά"))

        # Epiphany.
        self._add_epiphany_day(tr("Θεοφάνεια"), GREGORIAN_CALENDAR)

        # Clean Monday.
        self._add_ash_monday(tr("Καθαρά Δευτέρα"))

        # Greek Independence Day.
        self._add_holiday(tr("Εικοστή Πέμπτη Μαρτίου"), MAR, 25)

        # Cyprus National Day.
        self._add_holiday(tr("1η Απριλίου"), APR, 1)

        # Good Friday.
        self._add_good_friday(tr("Μεγάλη Παρασκευή"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Κυριακή του Πάσχα"))

        # Easter Monday.
        self._add_easter_monday(tr("Δευτέρα του Πάσχα"))

        # Labour Day.
        self._add_labor_day(tr("Εργατική Πρωτομαγιά"))

        # Monday of the Holy Spirit.
        self._add_whit_monday(tr("Δευτέρα του Αγίου Πνεύματος"))

        # Assumption of Mary.
        self._add_assumption_of_mary_day(tr("Κοίμηση της Θεοτόκου"))

        # Cyprus Independence Day.
        self._add_holiday(tr("Ημέρα Ανεξαρτησίας της Κύπρου"), OCT, 1)

        # Ochi Day.
        self._add_holiday(tr("Ημέρα του Όχι"), OCT, 28)

        # Christmas Eve.
        self._add_christmas_eve(tr("Παραμονή Χριστουγέννων"), GREGORIAN_CALENDAR)

        # Christmas Day.
        self._add_christmas_day(tr("Χριστούγεννα"), GREGORIAN_CALENDAR)

        # Day After Christmas.
        self._add_christmas_day_two(tr("Δεύτερη μέρα Χριστουγέννων"), GREGORIAN_CALENDAR)


class CY(Cyprus):
    pass


class CYP(Cyprus):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from gettext import gettext as tr

from holidays.calendars.gregorian import APR, JUL, AUG, SEP, DEC, MON, SUN
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class CostaRica(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_Costa_Rica
    - Law #8442 from 19.04.2005
    - Law #8604 from 17.09.2007
    - Law #8753 from 25.07.2009
    - Law #8886 from 01.11.2010
    - Law #9803 from 19.05.2020
    - Law #10050 from 25.10.2021
    """

    country = "CR"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _move_holiday(self, dt: date, forward: bool = False) -> None:
        if not self.observed or self._is_monday(dt) or (forward and self._is_weekend(dt)):
            return None
        obs_dt = (
            self._get_nth_weekday_from(1, MON, dt)
            if forward or not (self._is_tuesday(dt) or self._is_wednesday(dt))
            else self._get_nth_weekday_from(-1, MON, dt)
        )
        self._add_holiday(self.tr("%s (Observado)") % self[dt], obs_dt)
        self.pop(dt)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Maundy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Juan Santamaría Day.
        dt = self._add_holiday(tr("Día de Juan Santamaría"), APR, 11)
        if 2006 <= year <= 2010:
            self._move_holiday(dt, forward=True)
        elif year in {2023, 2024}:
            self._move_holiday(dt)

        # International Labor Day.
        dt = self._add_labor_day(tr("Día Internacional del Trabajo"))
        if year == 2021:
            self._move_holiday(dt)

        # Annexation of the Party of Nicoya to Costa Rica.
        dt = self._add_holiday(tr("Anexión del Partido de Nicoya a Costa Rica"), JUL, 25)
        if 2005 <= year <= 2008:
            self._move_holiday(dt, forward=True)
        elif 2020 <= year <= 2024:
            self._move_holiday(dt)

        # Feast of Our Lady of the Angels.
        self._add_holiday(tr("Fiesta de Nuestra Señora de los Ángeles"), AUG, 2)

        # Mother's Day.
        dt = self._add_assumption_of_mary_day(tr("Día de la Madre"))
        if 2005 <= year <= 2007:
            self._move_holiday(dt, forward=True)
        elif year in {2020, 2023, 2024}:
            self._move_holiday(dt)

        if year >= 2022:
            dt = date(year, AUG, 31)
            # Day of the Black Person and Afro-Costa Rican Culture.
            name = self.tr("Día de la Persona Negra y la Cultura Afrocostarricense")
            if self.observed and year in {2022, 2023}:
                dt = self._get_nth_weekday_from(1, SUN, dt)
                name = self.tr("%s (Observado)") % name
            self._add_holiday(name, dt)

        # Independence Day.
        dt = self._add_holiday(tr("Día de la Independencia"), SEP, 15)
        if year in {2020, 2021, 2022, 2024}:
            self._move_holiday(dt)

        if year <= 2019:
            # Cultures Day.
            self._move_holiday(self._add_columbus_day(tr("Día de las Culturas")), forward=True)

        if year >= 2020:
            # Army Abolition Day.
            dt = self._add_holiday(tr("Día de la Abolición del Ejército"), DEC, 1)
            if year in {2020, 2021, 2022}:
                self._move_holiday(dt)

        # Christmas Day.
        self._add_christmas_day(tr("Navidad"))


class CR(CostaRica):
    pass


class CRI(CostaRica):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr
from typing import Tuple

from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    MON,
)
from holidays.helpers import _normalize_tuple
from holidays.holiday_base import HolidayBase


class Japan(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Japan
    """

    country = "JP"
    default_language = "ja"
    special_holidays = {
        1959: (APR, 10, tr("結婚の儀")),  # The Crown Prince marriage ceremony.
        1989: (FEB, 24, tr("大喪の礼")),  # State Funeral of Emperor Shōwa.
        1990: (NOV, 12, tr("即位礼正殿の儀")),  # Enthronement ceremony.
        1993: (JUN, 9, tr("結婚の儀")),  # The Crown Prince marriage ceremony.
        2019: (
            (MAY, 1, tr("天皇の即位の日")),  # Enthronement day.
            (OCT, 22, tr("即位礼正殿の儀が行われる日")),  # Enthronement ceremony.
        ),
    }
    supported_languages = ("en_US", "ja")

    def _populate(self, year):
        if year < 1949 or year > 2099:
            raise NotImplementedError

        super()._populate(year)
        observed_dates = set()

        # New Year's Day.
        observed_dates.add(self._add_holiday(tr("元日"), JAN, 1))

        dt = date(year, JAN, 15) if year <= 1999 else self._get_nth_weekday_of_month(2, MON, JAN)
        # Coming of Age Day.
        observed_dates.add(self._add_holiday(tr("成人の日"), dt))

        if year >= 1967:
            # Foundation Day.
            observed_dates.add(self._add_holiday(tr("建国記念の日"), FEB, 11))

        if year >= 2020:
            # Emperor's Birthday.
            observed_dates.add(self._add_holiday(tr("天皇誕生日"), FEB, 23))

        # Vernal Equinox Day.
        observed_dates.add(self._add_holiday(tr("春分の日"), *self._vernal_equinox_date))

        # Showa Emperor's Birthday, Greenery Day or Showa Day.
        if year <= 1988:
            name = tr("天皇誕生日")
        elif year <= 2006:
            # Greenery Day.
            name = tr("みどりの日")
        else:
            # Showa Day.
            name = tr("昭和の日")
        observed_dates.add(self._add_holiday(name, APR, 29))

        # Constitution Day.
        observed_dates.add(self._add_holiday(tr("憲法記念日"), MAY, 3))

        # Greenery Day.
        if year >= 2007:
            observed_dates.add(self._add_holiday(tr("みどりの日"), MAY, 4))

        # Children's Day.
        observed_dates.add(self._add_holiday(tr("こどもの日"), MAY, 5))

        if year >= 1996:
            if year <= 2002:
                dt = date(year, JUL, 20)
            else:
                dates = {
                    2020: date(2020, JUL, 23),
                    2021: date(2021, JUL, 22),
                }
                dt = dates.get(year, self._get_nth_weekday_of_month(3, MON, JUL))
            # Marine Day.
            observed_dates.add(self._add_holiday(tr("海の日"), dt))

        if year >= 2016:
            dates = {
                2020: date(2020, AUG, 10),
                2021: date(2021, AUG, 8),
            }
            dt = dates.get(year, date(year, AUG, 11))
            # Mountain Day.
            observed_dates.add(self._add_holiday(tr("山の日"), dt))

        if year >= 1966:
            dt = (
                self._get_nth_weekday_of_month(3, MON, SEP)
                if year >= 2003
                else date(year, SEP, 15)
            )
            # Respect for the Aged Day.
            observed_dates.add(self._add_holiday(tr("敬老の日"), dt))

        # Autumnal Equinox Day.
        observed_dates.add(self._add_holiday(tr("秋分の日"), *self._autumnal_equinox_date))

        # Physical Education and Sports Day.
        if year >= 1966:
            name = (
                # Sports Day.
                tr("スポーツの日")
                if year >= 2020
                # Physical Education Day.
                else tr("体育の日")
            )
            dates = {
                2020: date(2020, JUL, 24),
                2021: date(2021, JUL, 23),
            }
            dt = dates.get(
                year,
                self._get_nth_weekday_of_month(2, MON, OCT)
                if year >= 2000
                else date(year, OCT, 10),
            )
            observed_dates.add(self._add_holiday(name, dt))

        # Culture Day.
        observed_dates.add(self._add_holiday(tr("文化の日"), NOV, 3))

        # Labor Thanksgiving Day.
        observed_dates.add(self._add_holiday(tr("勤労感謝の日"), NOV, 23))

        # Regarding the Emperor of Heisei.
        if 1989 <= year <= 2018:
            observed_dates.add(self._add_holiday(tr("天皇誕生日"), DEC, 23))

        if self.observed:
            for month, day, _ in _normalize_tuple(self.special_holidays.get(year, ())):
                observed_dates.add(date(year, month, day))

            # When a national holiday falls on Sunday, next working day
            # shall become a public holiday (振替休日) - substitute holidays.
            for dt in observed_dates.copy():
                if not self._is_sunday(dt):
                    continue
                hol_date = dt + td(days=+1)
                while hol_date in observed_dates:
                    hol_date += td(days=+1)
                # Substitute Holiday.
                observed_dates.add(self._add_holiday(tr("振替休日"), hol_date))

            # A weekday between national holidays becomes
            # a holiday too (国民の休日) - citizens' holidays.
            for dt in observed_dates:
                if dt + td(days=+2) not in observed_dates:
                    continue
                hol_date = dt + td(days=+1)
                if self._is_sunday(hol_date) or hol_date in observed_dates:
                    continue
                # National Holiday.
                self._add_holiday(tr("国民の休日"), hol_date)

    @property
    def _vernal_equinox_date(self) -> Tuple[int, int]:
        day = 20
        if (
            (self._year % 4 == 0 and self._year <= 1956)
            or (self._year % 4 == 1 and self._year <= 1989)
            or (self._year % 4 == 2 and self._year <= 2022)
            or (self._year % 4 == 3 and self._year <= 2055)
        ):
            day = 21
        elif self._year % 4 == 0 and self._year >= 2092:
            day = 19
        return MAR, day

    @property
    def _autumnal_equinox_date(self) -> Tuple[int, int]:
        day = 23
        if self._year % 4 == 3 and self._year <= 1979:
            day = 24
        elif (
            (self._year % 4 == 0 and self._year >= 2012)
            or (self._year % 4 == 1 and self._year >= 2045)
            or (self._year % 4 == 2 and self._year >= 2078)
        ):
            day = 22
        return SEP, day


class JP(Japan):
    pass


class JPN(Japan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import MAR, APR, JUL, AUG, SEP, OCT, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Zambia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.officeholidays.com/countries/zambia/
    https://www.timeanddate.com/holidays/zambia/
    https://en.wikipedia.org/wiki/Public_holidays_in_Zambia
    https://www.parliament.gov.zm/sites/default/files/documents/acts/Public%20Holidays%20Act.pdf
    """

    country = "ZM"
    special_holidays = {
        2016: (
            (AUG, 11, "General elections and referendum"),
            (SEP, 13, "Inauguration ceremony of President-elect and Vice President-elect"),
        ),
        2018: (
            (MAR, 9, "Public holiday"),
            (JUL, 26, "Lusaka mayoral and other local government elections"),
        ),
        2021: (
            (JUL, 2, "Memorial service for Kenneth Kaunda"),
            (JUL, 7, "Funeral of Kenneth Kaunda"),
            (AUG, 12, "General elections"),
            (AUG, 13, "Counting in general elections"),
            (AUG, 24, "Presidential inauguration"),
        ),
        2022: (MAR, 18, "Funeral of Rupiah Banda"),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date) -> None:
        # whenever a public holiday falls on a Sunday,
        # it rolls over to the following Monday
        if self.observed and self._is_sunday(dt):
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=+1))

    def _populate(self, year):
        # Observed since 1965
        if year <= 1964:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_observed(self._add_new_years_day("New Year's Day"))

        if year >= 1991:
            self._add_observed(
                # International Women's Day.
                self._add_womens_day("International Women's Day")
            )

        # Youth Day.
        self._add_observed(self._add_holiday("Youth Day", MAR, 12))

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Holy Saturday.
        self._add_holy_saturday("Holy Saturday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        if year >= 2022:
            # Kenneth Kaunda Day.
            self._add_observed(self._add_holiday("Kenneth Kaunda Day", APR, 28))

        # Labour Day.
        self._add_observed(self._add_labor_day("Labour Day"))

        # Africa Freedom Day.
        self._add_observed(self._add_africa_day("Africa Freedom Day"))

        first_mon_of_july = self._add_holiday(
            # Heroes' Day.
            "Heroes' Day",
            self._get_nth_weekday_of_month(1, MON, JUL),
        )

        # Unity Day.
        self._add_holiday("Unity Day", first_mon_of_july + td(days=+1))

        # Farmers' Day.
        self._add_holiday("Farmers' Day", self._get_nth_weekday_of_month(1, MON, AUG))

        if year >= 2015:
            # National Prayer Day.
            self._add_observed(self._add_holiday("National Prayer Day", OCT, 18))

        # Independence Day.
        self._add_observed(self._add_holiday("Independence Day", OCT, 24))

        # Christmas Day.
        self._add_observed(self._add_christmas_day("Christmas Day"))


class ZM(Zambia):
    pass


class ZMB(Zambia):
    pass
